<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Version details
 *
 * @package    block_iclicker
 * @copyright  2012 i>clicker (R) <http://www.iclicker.com/dnn/> onwards Saji Nair (http://sajinair.com)
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

require_once(dirname(__FILE__).'/../../config.php');
global $CFG, $USER, $COURSE;
// link in external libraries
// @noinspection PhpIncludeInspection */
require_once($CFG->libdir.'/gradelib.php');
// @noinspection PhpIncludeInspection */
require_once($CFG->libdir.'/dmllib.php');
// @noinspection PhpIncludeInspection */
require_once($CFG->libdir.'/accesslib.php');

/**
 * For XML error handling
 * @param $errno
 * @param $errstr
 * @param $errfile
 * @param $errline
 * @return bool
 * @throws DOMException
 */
function iclicker_handlexmlerror($errno, $errstr,  /* * @noinspection PhpUnusedParameterInspection */
                                 $errfile,  /* * @noinspection PhpUnusedParameterInspection */
                                 $errline) {
    if ($errno == E_WARNING && (substr_count($errstr, "DOMDocument::loadXML()") > 0)) {
        throw new DOMException($errstr);
    } else {
        return false;
    }
}

/**
 * Defines an exception which can occur when validating clicker ids
 * Valid types are:
 * empty - the clickerId is null or empty string
 * length - the clickerId length is not 8 chars (too long), shorter clickerIds are padded out to 8
 * chars - the clickerId contains invalid characters
 * checksum - the clickerId did not validate using the checksum method
 * sample - the clickerId matches the sample one and cannot be used
 */
class ClickerIdInvalidException extends Exception {
    const F_EMPTY = 'EMPTY';
    const F_LENGTH = 'LENGTH';
    const F_CHARS = 'CHARS';
    const GO_CHARS = 'GO_CHARS';
    const F_CHECKSUM = 'CHECKSUM';
    const F_SAMPLE = 'SAMPLE';
    const GO_NO_USER = 'GO_NO_USER';
    const GO_LASTNAME = 'GO_LASTNAME';
    const GO_NO_MATCH = 'GO_NO_MATCH';
    public $type = 'UNKNOWN';
    public $clickerid = null;
    /**
     * @param string $message the error message
     * @param string $type [optional] Valid types are:
     * empty - the clickerId is null or empty string
     * length - the clickerId length is not 8 chars (too long), shorter clickerIds are padded out to 8
     * chars - the clickerId contains invalid characters
     * checksum - the clickerId did not validate using the checksum method
     * sample - the clickerId matches the sample one and cannot be used
     * @param string $clickerid [optional] the clicker id
     */
    public function __construct($message, $type = null, $clickerid = null) {
        parent::__construct($message);
        $this->type = $type;
        $this->clickerid = $clickerid;
    }
    public function error_message() {
        $errormsg = 'Error on line '.$this->getLine().' in '.$this->getFile().': '.$this->getMessage().' : type='.$this->type.' : clickerid='.$this->clickerid;
        return $errormsg;
    }
}

class ClickerRegisteredException extends Exception {
    public $ownerid;
    public $clickerid;
    public $registeredownerid;
    public function __construct($message, $ownerid, $clickerid, $registeredownerid) {
        parent::__construct($message);
        $this->ownerid = $ownerid;
        $this->clickerid = $clickerid;
        $this->registeredownerid = $registeredownerid;
    }
    public function error_message() {
        $errormsg = 'Error on line '.$this->getLine().' in '.$this->getFile().': '.$this->getMessage().' : cannot register to '.$this->ownerid;
        $errormsg .= ', clicker already registered to owner='.$this->registeredownerid.' : clickerid='.$this->clickerid;
        return $errormsg;
    }
}

/**
 * Indicates that this connection requires SSL
 */
class ClickerSSLRequiredException extends Exception {
}

/**
 * This marks an exception as being related to an authn or authz failure
 */
class ClickerSecurityException extends Exception {
}

class ClickerWebservicesException extends Exception {
}

/**
 * This holds all the service logic for the iclicker integrate plugin
 */
class iclicker_service {

    // CONSTANTS
    const VERSION = '2.2'; // MUST match version.php
    const BLOCK_VERSION = 2017080200; // MUST match version.php

    // Moodle version - 2.0 = 2010112400; 2.1 = 2011070100; 2.2 = 2011120100; 2.3 = 2012062500; 2.4 = 2012120300

    const BLOCK_NAME = 'block_iclicker';
    const BLOCK_PATH = '/blocks/iclicker';
    const REG_TABLENAME = 'iclicker_registration';
    const REG_ORDER = 'timecreated desc';
    const USER_KEY_TABLENAME = 'iclicker_user_key';
    const GRADE_CATEGORY_NAME = 'iclicker polling scores'; // default category
    const GRADE_ITEM_TYPE = 'manual';
    const GRADE_ITEM_MODULE = null; // 'iclicker'; - Moodle 2.8 fix
    const GRADE_LOCATION_STR = 'manual';
    const DEFAULT_SYNC_HOUR = 3;
    const BLOCK_RUNNER_KEY = 'block_iclicker_runner';
    const DEFAULT_SERVER_URL = "http://moodle.org/"; // "http://epicurus.learningmate.com/";

    const NATIONAL_WS_URL = "https://webservices.iclicker.com/iclicker_gbsync_registrations/service.asmx";
    /*
     * iclicker_gbsync_reg / #8d7608e1e7f4@
     * 'Basic ' + base64(username + ":" + password)
     */
    const NATIONAL_WS_BASIC_AUTH_HEADER = 'Basic aWNsaWNrZXJfZ2JzeW5jX3JlZzojOGQ3NjA4ZTFlN2Y0QA==';
    const NATIONAL_WS_AUTH_USERNAME = 'iclicker_gbsync_reg';
    const NATIONAL_WS_AUTH_PASSWORD = '#8d7608e1e7f4@';

    // errors constants
    const SCORE_UPDATE_ERRORS = 'ScoreUpdateErrors';
    const POINTS_POSSIBLE_UPDATE_ERRORS = 'PointsPossibleUpdateErrors';
    const USER_DOES_NOT_EXIST_ERROR = 'UserDoesNotExistError';
    const GENERAL_ERRORS = 'GeneralErrors';
    const SCORE_KEY = '${SCORE}';

    // CLASS VARIABLES

    // CONFIG
    public static $serverurl = self::DEFAULT_SERVER_URL;
    public static $domainurl = self::DEFAULT_SERVER_URL;
    public static $allowremotesharing = true; // FORCE this to true
    public static $disablealternateid = false;
    public static $usenationalwebservices = false;
    public static $webservicesurl = self::NATIONAL_WS_URL;
    public static $webservicesusername = self::NATIONAL_WS_AUTH_USERNAME;
    public static $webservicespassword = self::NATIONAL_WS_AUTH_PASSWORD;
    public static $maxcoursestofetch = 100;
    public static $disablesyncwithnational = false;
    public static $blockiclickerssoenabled = false;
    public static $blockiclickerssosharedkey = null;
    public static $enablecourseshortname = false;
    public static $testmode = false;


    // STATIC METHODS

    /**
     * @static
     * @param string $added extra path to add
     * @return string the path for this block
     */
    public static function block_path($added = null) {
        global $CFG;
        if (isset($added)) {
            $added = '/'.$added;
        } else {
            $added = '';
        }
        return $CFG->dirroot.self::BLOCK_PATH.$added;
    }

    /**
     * @static
     * @param string $added extra path to add
     * @return string url for this block
     */
    public static function block_url($added = null) {
        global $CFG;
        if (isset($added)) {
            $added = '/'.$added;
        } else {
            $added = '';
        }
        return $CFG->wwwroot.self::BLOCK_PATH.$added;
    }

    /**
     * i18n message handling
     *
     * @param string $key i18 msg key
     * @param object $vars [optional] optional replacement variables
     * @return string the translated string
     */
    public static function msg($key, $vars = null) {
        return get_string($key, self::BLOCK_NAME, $vars);
    }

    public static function df($time) {
        return userdate($time, get_string('strftimedaydate', 'langconfig')); // strftime('%Y/%m/%d', $time); //userdate($time, '%Y/%m/%d');
    }

    /**
     * Sends an email to an email address
     *
     * @param string $to email address to send email to
     * @param string $subject email subject
     * @param string $body email body
     * @return bool true if email sent, false otherwise
     */
    public static function send_email($to, $subject, $body) {
        // $user should be a fake user object with the email set to the correct value, $from should be a string
        $user = new stdClass();
        $user->email = $to;
        $user->firstname = 'ADMIN';
        $user->lastname = $to;
        $user->mailformat = 0; // plain
        $user->confirmed = 1;
        $user->deleted = 0;
        $user->emailstop = 0;
        $user->id = 1;
        if (email_to_user($user, "SYSTEM", $subject, $body) !== true) {
            debugging("Could not send email ($to) : $subject \n $body");
        }
    }

    /**
     * Sends a notification to the configured email addresses in the system about a failure
     *
     * @param string $message the message to send
     * @param object $exception [optional] the optional exception to notify the admins about
     * @return bool true if email sent, false otherwise
     */
    public static function send_notifications($message, $exception = null) {
        global $CFG;
        // load these on demand only - block_iclicker_notify_emails
        $adminemails = null;
        if (!empty($CFG->block_iclicker_notify_emails)) {
            $emailstring = $CFG->block_iclicker_notify_emails;
            $adminemails = explode(',', $emailstring);
            array_walk($adminemails, 'trim');
        }

        // add to failures record and trim it to 5
        $failures = self::get_failures();
        $msg = $message;
        if ($exception != null) {
            $msg .= " Failure: ".$exception->message." ".$exception;
        }
        array_unshift($failures, date('Y-m-d h:i:s').' :: '.substr($msg, 0, 300));
        while (count($failures) > 5) {
            array_pop($failures);
        }
        set_config('block_iclicker_failures', implode('*****', $failures), self::BLOCK_NAME);

        $body = "i>clicker Moodle integrate plugin notification (".date('d.m.Y h:i:s').")\n" . $message . "\n";
        if ($exception != null) {
            $body .= "\nFailure:\n".$exception->message."\n\n".$exception;
        }
        if (!empty($adminemails)) {
            foreach ($adminemails as $email) {
                self::send_email($email, 'i>clicker Moodle integrate plugin notification', $body);
            }
            $sent = true;
        } else {
            debugging("No emails set for sending notifications: logging notification: $body");
            $sent = false;
        }
        return $sent;
    }

    public static function get_failures() {
        $failures = array();
        $failurestring = get_config(self::BLOCK_NAME, 'block_iclicker_failures');
        if (! empty($failurestring)) {
            $failures = explode('*****', $failurestring);
        }
        return $failures;
    }

    // USERS
    const USER_FIELDS = 'id,username,picture,firstname,lastname,firstnamephonetic,lastnamephonetic,middlename,alternatename,imagealt,email';

    /**
     * Authenticate a user by username and password
     * @static
     * @param $username
     * @param $password
     * @param string $ssokey [OPTIONAL] the sso key sent in the request OR null if there was not one
     * @return bool true if the authentication is successful
     * @throws ClickerSecurityException  if auth invalid
     */
    public static function authenticate_user($username, $password, $ssokey = null) {
        global $USER;
        if (!isset($USER->id) || !$USER->id) {
            // this user is not already logged in
            if (self::$blockiclickerssoenabled) {
                self::verify_key($ssokey); // verify the key is valid
                $user = self::get_user_by_username($username, true);
                if ($user && self::check_user_key($user->id, $password)) {
                    complete_user_login($user);
                } else {
                    throw new ClickerSecurityException('Could not SSO authenticate username ('.$username.')');
                }
            } else {
                $u = authenticate_user_login($username, $password);
                if ($u === false) {
                    throw new ClickerSecurityException('Could not authenticate username ('.$username.')');
                }
                complete_user_login($u);
            }
        }
        return true;
    }

    /**
     * Ensure user is logged in and return the current user id
     * @return string the current user id
     * @throws ClickerSecurityException if there is no current user
     * @static
     */
    public static function require_user() {
        global $USER;
        if (!isset($USER->id) || !$USER->id) {
            throw new ClickerSecurityException('User must be logged in');
        } else if (isguestuser($USER->id)) {
            throw new ClickerSecurityException('Guest users ('.$USER->id.') are not allowed to access the clicker tool');
        }
        return $USER->id;
    }

    /**
     * Gets the current userid, return false if none can be found
     * @return int|boolean the current user id OR null/false if no user
     */
    public static function get_current_user_id() {
        $currentuser = null;
        try {
            $currentuser = self::require_user();
        } catch (ClickerSecurityException $e) {
            $currentuser = false;
        }
        return $currentuser;
    }

    /**
     * Gets a user by their username
     * @static
     * @param string $username the username (i.e. login name)
     * @param bool $allfields default false, if true then return all user fields, otherwise just the ones needed for iclicker handling
     * @return stdClass|bool the user object OR false if none can be found
     */
    public static function get_user_by_username($username, $allfields = false) {
        global $DB;
        $user = false;
        if ($username) {
            $fields = self::USER_FIELDS;
            if ($allfields) {
                $fields = '*';
            }
            $user = $DB->get_record('user', array('username' => $username), $fields);
            // TESTING handling
            if (self::$testmode && !$user) {
                // test users
                if ($username == 'student01') {
                    $user = new stdClass();
                    $user->id = 101;
                    $user->username = 'student01';
                    $user->firstname = 'Student';
                    $user->lastname = 'One';
                    $user->email = 'one@fail.com';
                    $user->firstnamephonetic = 'student';
                    $user->lastnamephonetic = 'one';
                    $user->middlename = 's';
                    $user->alternatename = 'student01';
                } else if ($username == 'student02') {
                    $user = new stdClass();
                    $user->id = 102;
                    $user->username = 'student02';
                    $user->firstname = 'Student';
                    $user->lastname = 'Two';
                    $user->email = 'two@fail.com';
                    $user->firstnamephonetic = 'Student';
                    $user->lastnamephonetic = 'Two';
                    $user->middlename = 's';
                    $user->alternatename = 'student02';
                } else if ($username == 'student03') {
                    $user = new stdClass();
                    $user->id = 103;
                    $user->username = 'student03';
                    $user->firstname = 'Student';
                    $user->lastname = 'Three';
                    $user->email = 'three@fail.com';
                    $user->firstnamephonetic = 'Student';
                    $user->lastnamephonetic = 'Three';
                    $user->middlename = 's';
                    $user->alternatename = 'student03';
                } else if ($username == 'inst01') {
                    $user = new stdClass();
                    $user->id = 111;
                    $user->username = 'inst01';
                    $user->firstname = 'Instructor';
                    $user->lastname = 'One';
                    $user->email = 'uno_inst@fail.com';
                    $user->firstnamephonetic = 'Instructor';
                    $user->lastnamephonetic = 'One';
                    $user->middlename = 'I';
                    $user->alternatename = 'inst01';
                }
            }
        }
        return $user;
    }

    /**
     * Get user records for a set of user ids
     * @param array|int $userids an array of user ids OR a single userid
     * @return array|stdClass a map of userid -> user data OR single user object for single userid OR empty array if no matches
     */
    public static function get_users($userids) {
        global $DB;
        $results = array(
        );
        if (isset($userids)) {
            if (is_array($userids)) {
                $users = false;
                if (! empty($userids)) {
                    // $ids = implode(',', $userids);
                    $users = $DB->get_records_list('user', 'id', $userids, 'id', self::USER_FIELDS);
                }
                if ($users) {
                    foreach ($users as $user) {
                        self::make_user_display_name($user);
                        $results[$user->id] = $user;
                    }
                }
            } else {
                // Single user id.
                $user = $DB->get_record('user', array('id' => $userids), self::USER_FIELDS);
                // TESTING handling.
                if (self::$testmode && !$user) {
                    if ($userids == 101) {
                        $user = new stdClass();
                        $user->id = 101;
                        $user->username = 'student01';
                        $user->firstname = 'Student';
                        $user->lastname = 'One';
                        $user->email = 'one@fail.com';
                        $user->firstnamephonetic = 'student';
                        $user->lastnamephonetic = 'one';
                        $user->middlename = 's';
                        $user->alternatename = 'student01';
                    } else if ($userids == 102) {
                        $user = new stdClass();
                        $user->id = 102;
                        $user->username = 'student02';
                        $user->firstname = 'Student';
                        $user->lastname = 'Two';
                        $user->email = 'two@fail.com';
                        $user->firstnamephonetic = 'Student';
                        $user->lastnamephonetic = 'Two';
                        $user->middlename = 's';
                        $user->alternatename = 'student02';
                    } else if ($userids == 103) {
                        $user = new stdClass();
                        $user->id = 103;
                        $user->username = 'student03';
                        $user->firstname = 'Student';
                        $user->lastname = 'Three';
                        $user->email = 'three@fail.com';
                        $user->firstnamephonetic = 'Student';
                        $user->lastnamephonetic = 'Three';
                        $user->middlename = 's';
                        $user->alternatename = 'student03';
                    } else if ($userids == 111) {
                        $user = new stdClass();
                        $user->id = 111;
                        $user->username = 'inst01';
                        $user->firstname = 'Instructor';
                        $user->lastname = 'One';
                        $user->email = 'uno_inst@fail.com';
                        $user->firstnamephonetic = 'Instructor';
                        $user->lastnamephonetic = 'One';
                        $user->middlename = 'I';
                        $user->alternatename = 'inst01';
                    }
                }
                if ($user) {
                    self::make_user_display_name($user);
                    $results = $user;
                }
            }
        }
        return $results;
    }

    /**
     * Get a display name for a given user id
     * @param int $userid id for a user
     * @return string the display name
     */
    public static function get_user_displayname($userid) {
        $name = "UNKNOWN-".$userid;
        $user = self::get_users($userid);
        if ($user && isset($user->id)) {
            if (isset($user->display_name)) {
                $name = $user->display_name;
            } else {
                $name = self::make_user_display_name($user->id);
            }
        }
        return $name;
    }

    private static function make_user_display_name(&$user) {
        $displayname = fullname($user);
        $user->name = $displayname;
        $user->display_name = $displayname;
        return $displayname;
    }

    /**
     * @param int $userid [optional] the user id
     * @return bool true if this user is an admin OR false if not
     * @static
     */
    public static function is_admin($userid = null) {
        if (!isset($userid)) {
            try {
                $userid = self::require_user();
            } catch (ClickerSecurityException $e) {
                return false;
            }
        }
        $result = is_siteadmin($userid);
        return $result;
    }

    /**
     * Check if a user is an instructor in moodle
     *
     * @param int $userid [optional] the user id to check (default to current user)
     * @return bool true if an instructor or false otherwise
     * @static
     */
    public static function is_instructor($userid = null) {
        if (!isset($userid)) {
            try {
                $userid = self::require_user();
            } catch (ClickerSecurityException $e) {
                return false;
            }
        }

        $result = false;
        if (class_exists('context_course')) {
            // for Moodle 2.2+
            $courses = enrol_get_users_courses($userid, true, null, null);
            foreach ($courses as $id => $course) {
                $context = context_course::instance($id);
                if (has_capability('moodle/course:update', $context, $userid)) {
                    // unset($courses[$id]);
                    $result = true;
                    break;
                }
            }
        } else {
            // for Moodle before 2.2
            global $USER;
            // sadly this is the only way to do this check: http://moodle.org/mod/forum/discuss.php?d=140383
            $accessinfo = null;
            if ($userid === $USER->id && isset($USER->access)) {
                $accessinfo = $USER->access;
            } else {
                $accessinfo = get_user_access_sitewide($userid);
            }
            $results = enrol_get_users_courses($userid, true, array(), 'c.sortorder');
            $result = count($results) > 0;
        }
        return $result;
    }

    const CLICKERID_SAMPLE = '11A4C277';
    /**
     * Cleans up and validates a given clickerid
     *
     * @static
     * @param string $clickerid a remote clicker ID
     * @param string $userlastname [OPTIONAL] the last name of the user, only applies for GO IDs
     * @return string the cleaned up and valid clicker ID
     * @throws ClickerIdInvalidException if the id is invalid for some reason,
     * the exception will indicate the type of validation failure
     */
    public static function validate_clicker_id($clickerid, $userlastname = null) {
        global $USER;
        if (!isset($clickerid) || strlen($clickerid) == 0) {
            throw new ClickerIdInvalidException("empty or null clickerid", ClickerIdInvalidException::F_EMPTY, $clickerid);
        }
        $clickeridlength = strlen($clickerid);
        /*
        https://jira-mnv.atlassian.net/browse/ICL-2603 :
        Taking out references to i>clicker GO, GO is going to see end of life mid Aug and we need to take out all references for it
        if ($clickeridlength == 12) {
            // support for new clicker go ids
            $clickerid = strtoupper(trim($clickerid));
            if (!preg_match('/^[0-9A-Z]+$/', $clickerid)) {
                throw new ClickerIdInvalidException("clickerid can only contains A-Z and 0-9", ClickerIdInvalidException::GO_CHARS, $clickerid);
            }
            if (!isset($userlastname) || empty($userlastname)) {
                // fetch last name from current user
                if (self::get_current_user_id() === false) {
                    throw new ClickerIdInvalidException("No current user available, cannot validate clickerid: $clickerid", ClickerIdInvalidException::GO_NO_USER, $clickerid);
                }
                $userlastname = $USER->lastname;
            }
            self::ws_go_verify_clickerid($clickerid, $userlastname); // ClickerIdInvalidException exception if invalid (or WS exception)

        } else
        */
        if ($clickeridlength <= 8) {
            // length <= 8, support for old clicker device ids
            $clickerid = strtoupper(trim($clickerid));
            if (!preg_match('/^[0-9A-F]+$/', $clickerid)) {
                throw new ClickerIdInvalidException("clickerid can only contains A-F and 0-9", ClickerIdInvalidException::F_CHARS, $clickerid);
            }
            while (strlen($clickerid) < 8) {
                // front pad with zeros
                $clickerid = "0" . $clickerid;
            }
            if (self::CLICKERID_SAMPLE == $clickerid) {
                throw new ClickerIdInvalidException("clickerid cannot match the sample ID", ClickerIdInvalidException::F_SAMPLE, $clickerid);
            }
            $idarray = array();
            $idarray[0] = substr($clickerid, 0, 2);
            $idarray[1] = substr($clickerid, 2, 2);
            $idarray[2] = substr($clickerid, 4, 2);
            $idarray[3] = substr($clickerid, 6, 2);
            $checksum = 0;
            foreach ($idarray as $piece) {
                $hex = hexdec($piece);
                $checksum = $checksum ^ $hex;
            }
            if ($checksum != 0) {
                throw new ClickerIdInvalidException("clickerid checksum (" . $checksum . ") validation failed", ClickerIdInvalidException::F_CHECKSUM, $clickerid);
            }

        } else {
            // totally invalid clicker length
            throw new ClickerIdInvalidException("clickerid is an invalid length ($clickeridlength), must be 8 or 12 chars", ClickerIdInvalidException::F_LENGTH, $clickerid);
        }
        return $clickerid;
    }

    /**
     * For all remoteids starting with “2”, “4”, “8” we have to generate an alternate id
     * and concatenate it with the existing remote ids for that particular user in the data
     * sent to the iclicker desktop app (this is like creating an extra clickerid based on the existing ones)
     *
     * @static
     * @param string $clickerid a remote clicker ID
     * @return string|null a translated clicker ID OR null if no translation is required or id is invalid
     */
    public static function translate_clicker_id($clickerid) {
        $alternateid = null;
        try {
            // validate the input, do nothing but return null if invalid
            $clickerid = self::validate_clicker_id($clickerid);
            $startswith = $clickerid{0};
            if ('2' == $startswith || '4' == $startswith || '8' == $startswith) {
                // found clicker to translate
                $p1 = hexdec('0'.$clickerid{1});
                $p2 = hexdec(substr($clickerid, 2, 2));
                $p3 = hexdec(substr($clickerid, 4, 2));
                $p4 = $p1 ^ $p2 ^ $p3;
                $part4 = strtoupper(dechex($p4));
                $part4 = (strlen($part4) === 1 ? '0'.$part4 : $part4);
                $alternateid = '0' . substr($clickerid, 1, 5) . $part4;
            }
        } catch (ClickerIdInvalidException $e) {
            $alternateid = null;
        }
        return $alternateid;

    }


    // *******************************************************************************
    // SSO handling

    /**
     * Make or find a user key for the given user,
     * if they don't have one, this will create one, if they do it will retrieve it.
     * It can also be used to generate a new user key
     *
     * @static
     * @param string $userid [OPTIONAL] the internal user id (if null, use the current user id)
     * @param bool $makenew if true, make a new key even if the user already has one, if false, only make a key if they do not have one (default: false)
     * @return string the user key for the given user
     * @throws InvalidArgumentException if user is not set or is not an instructor
     */
    public static function make_user_key($userid = null, $makenew = false) {
        global $DB;
        if (!$userid) {
            $userid = self::get_current_user_id();
        }
        if (!$userid) {
            throw new InvalidArgumentException("no current user, cannot generate a user key");
        }
        if (!self::is_admin($userid)
                && !self::is_instructor($userid)
                ) {
            // if user is not an instructor or an admin then we will not make a key for them, this is to block students from getting a pass key
            throw new InvalidArgumentException("current user ($userid) is not an instructor, cannot generate user key for them");
        }
        // find the key for this user if one exists
        $userkey = null;
        $cuk = $DB->get_record(self::USER_KEY_TABLENAME, array('userid' => $userid), 'id, userkey');
        if ($makenew && $cuk !== false) {
            // remove the existing key so we can make a new one
            $DB->delete_records(self::USER_KEY_TABLENAME, array('id' => $cuk->id));
            $cuk = null;
        }
        if ($cuk == null) {
            // make a new key and store it
            $newkeyvalue = self::random_alpha_numeric(12);
            $cuk = new stdClass();
            $cuk->userid = $userid;
            $cuk->userkey = $newkeyvalue;
            try {
                $DB->insert_record(self::USER_KEY_TABLENAME, $cuk);
                $userkey = $newkeyvalue;
            } catch (dml_exception $e) {
                // this should not have happened but it means the key already exists somehow, probably a sync issue of some kind
                debugging("Failed when attempting to create a new clicker user key for :".$userid);
            }
        } else {
            $userkey = $cuk->userkey;
        }
        return $userkey;
    }

    /**
     * @static
     * @param int $length length of the random string
     * @return string random alphanumeric string
     */
    public static function random_alpha_numeric($length) {
        $characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
        $characterslength = strlen($characters);
        $string = '';
        for ($i = 0; $i < $length; $i++) {
            $string .= $characters[rand(0, $characterslength - 1)];
        }
        return $string;
    }

    /**
     * Find the user key for a given user if one exists
     *
     * @param string $userid [OPTIONAL] the internal user id (if null, use the current user id)
     * @return string the user key for the given user OR null if there is no key
     * @throws InvalidArgumentException
     */
    public static function get_user_key($userid) {
        global $DB;
        if ($userid == null) {
            $userid = self::get_current_user_id();
        }
        if ($userid == null) {
            throw new InvalidArgumentException("no current user, cannot check user key");
        }
        $key = null;
        $cuk = $DB->get_record(self::USER_KEY_TABLENAME, array('userid' => $userid), 'id,userkey');
        if ($cuk !== false) {
            $key = $cuk->userkey;
        }
        return $key;
    }

    /**
     * Checks if the passed in user key is valid compared to the internally stored user key
     *
     * @param string $userid [OPTIONAL] the internal user id (if null, use the current user id)
     * @param string $userkey the passed in SSO key to check for this user
     * @return bool true if the key is valid OR false if the user has no key or the key is otherwise invalid
     * @throws InvalidArgumentException
     */
    public static function check_user_key($userid, $userkey) {
        global $DB;
        if (empty($userkey)) {
            throw new InvalidArgumentException("userKey cannot be empty");
        }
        if ($userid == null) {
            $userid = self::get_current_user_id();
        }
        if ($userid == null) {
            throw new InvalidArgumentException("no current user, cannot check user key");
        }
        $valid = false;
        $cuk = $DB->get_record(self::USER_KEY_TABLENAME, array('userid' => $userid), 'id,userkey');
        if ($cuk !== false) {
            if ($userkey == $cuk->userkey) {
                $valid = true;
            }
        }
        return $valid;
    }

    /**
     * @param string $sharedkey set and verify the shared key (if invalid, log a warning)
     */
    public static function set_shared_key($sharedkey) {
        if ($sharedkey != null) {
            if (strlen($sharedkey) < 10) {
                debugging("i>clicker shared key ($sharedkey) is too short, must be at least 10 chars long. SSO shared key will be ignored until a longer key is entered.");
            } else {
                self::$blockiclickerssoenabled = true;
                self::$blockiclickerssosharedkey = $sharedkey;
                if (!self::$testmode) {
                    debugging("i>clicker plugin SSO handling enabled by shared key, note that this will disable normal username/password handling");
                }
            }
        }
    }

    /**
     * @return string the SSO shared key value (or empty string otherwise)
     * NOTE: this only works if SSO is enabled AND the user is an admin
     */
    public static function get_shared_key() {
        $key = '';
        if (self::$blockiclickerssoenabled && self::is_admin()) {
            $key = self::$blockiclickerssosharedkey;
        }
        return $key;
    }

    /**
     * @return bool true if SSO handling is enabled, false otherwise
     */
    public static function is_single_sign_on_enabled() {
        return self::$blockiclickerssoenabled;
    }

    /**
     * Verify the passed in encrypted SSO shared key is valid,
     * this will return false if the key is not configured
     *
     * Key must have been encoded like so (where timestamp is the unix time in seconds):
     * sentKey = hex(sha1(sharedKey + ":" + timestamp)) + "|" + timestamp
     *
     * @static
     * @param string $key the passed in key (should already be sha-1 and hex encoded with the timestamp appended)
     * @return bool true if the key is valid, false if SSO shared keys are disabled
     * @throws InvalidArgumentException if the key format is invalid
     * @throws ClickerSecurityException if the key timestamp has expired or the key does not match
     */
    public static function verify_key($key) {
        if (empty($key)) {
            throw new InvalidArgumentException("key must be set in order to verify the key");
        }
        $verified = false;
        if (self::$blockiclickerssoenabled) {
            // encoding process requires the key and timestamp so split them from the passed in key
            $splitindex = strrpos($key, '|');
            if (($splitindex === -1) || (strlen($key) < $splitindex + 1)) {
                throw new InvalidArgumentException("i>clicker shared key format is invalid (no |), must be {encoded key}|{timestamp}");
            }
            $actualkey = substr($key, 0, $splitindex);
            if (empty($actualkey)) {
                throw new InvalidArgumentException("i>clicker shared key format is invalid (missing encoded key), must be {encoded key}|{timestamp}");
            }
            $timestampstr = substr($key, $splitindex + 1);
            if (empty($timestampstr)) {
                throw new InvalidArgumentException("i>clicker shared key format is invalid (missing timestamp), must be {encoded key}|{timestamp}");
            } else if (!is_numeric($timestampstr)) {
                throw new InvalidArgumentException("i>clicker shared key format is invalid (non numeric timestamp), must be {encoded key}|{timestamp}");
            }
            $timestamp = (int) $timestampstr;

            // check this key is still good (must be within 5 mins of now)
            $unixtime = time();
            $timediff = abs($timestamp - $unixtime);
            if ($timediff > 300) {
                $errmsg = "i>clicker shared key timestamp is out of date, this timestamp ($timestamp) is more than 5 minutes different from the current time ($unixtime)";
                throw new ClickerSecurityException($errmsg);
            }

            // finally we verify the key with the one in the config
            $sha1hex = self::make_encoded_key($timestamp);
            if ($actualkey !== $sha1hex) {
                throw new ClickerSecurityException("i>clicker encoded shared key ($actualkey) does not match with the key in Sakai");
            }
            $verified = true;
        }
        return $verified;
    }

    /**
     * Creates a key using the current $blockiclickerssosharedkey
     * @static
     * @param int $timestamp [OPTIONAL] the unix timestamp OR uses now if it is not set
     * @return string the encoded key OR null if SSO is disabled
     */
    public static function make_encoded_key($timestamp=-1) {
        $encodedkey = null;
        if (self::$blockiclickerssoenabled) {
            if ($timestamp < 1) {
                $timestamp = time();
            }
            $encodedkey = sha1(self::$blockiclickerssosharedkey . ":" . $timestamp);
        }
        return $encodedkey;
    }



    // CLICKER REGISTRATIONS DATA

    /**
     * @static
     * @param int $regid the registration ID
     * @return stdClass|bool the registration object OR false if none found
     * @throws InvalidArgumentException
     */
    public static function get_registration_by_id($regid) {
        global $DB;
        if (!isset($regid)) {
            throw new InvalidArgumentException("reg_id must be set");
        }
        $result = $DB->get_record(self::REG_TABLENAME, array('id' => $regid));
        return $result;
    }

    /**
     * @static
     * @param string $clickerid the clicker id
     * @param int $userid [optional] the user who registered the clicker (id)
     * @return stdClass|bool the registration object OR false if none found
     * @throws InvalidArgumentException|ClickerSecurityException
     */
    public static function get_registration_by_clicker_id($clickerid, $userid = null) {
        global $DB;
        if (!$clickerid) {
            throw new InvalidArgumentException("clickerid must be set");
        }
        try {
            $clickerid = self::validate_clicker_id($clickerid);
        } catch (ClickerIdInvalidException $e) {
            return false;
        }
        // NOTE: also returns disabled registrations
        $params = array('clickerid' => $clickerid);
        if (!empty($userid)) {
            $params['ownerid'] = $userid;
        }
        $result = $DB->get_record(self::REG_TABLENAME, $params, $fields = '*', $strictness = IGNORE_MULTIPLE); // only get the first one
        if ($result && !empty($userid)) {
            $currentuserid = self::require_user();
            if (!self::can_read_registration($result, $currentuserid)) {
                throw new ClickerSecurityException("User ($currentuserid) not allowed to access registration ($result->id)");
            }
        }
        return $result;
    }

    public static function can_read_registration($clickerregistration, $userid) {
        if (!isset($clickerregistration)) {
            throw new InvalidArgumentException("clicker_registration must be set");
        }
        if (!isset($userid)) {
            throw new InvalidArgumentException("userid must be set");
        }
        $result = false;
        if ($clickerregistration->ownerid == $userid) {
            $result = true;
        } else if (self::is_admin($userid)) {
            $result = true;
        } else if (self::is_instructor($userid)) {
            // NOTE: simply allowing instructors to read/write all registrations for now
            $result = true;
        }
        return $result;
    }

    public static function can_write_registration($clickerregistration, $userid) {
        if (!isset($clickerregistration)) {
            throw new InvalidArgumentException("clicker_registration must be set");
        }
        if (!isset($userid)) {
            throw new InvalidArgumentException("userid must be set");
        }
        $result = false;
        if ($clickerregistration->ownerid == $userid) {
            $result = true;
        } else if (self::is_admin($userid)) {
            $result = true;
        } else if (self::is_instructor($userid)) {
            // NOTE: simply allowing instructors to read/write all registrations for now
            $result = true;
        }
        return $result;
    }

    /**
     * @param int $userid [optional] the user id OR current user id
     * @param boolean $activated if null or not set then return all,
     * if true then return active only, if false then return inactive only
     * @return array the list of registrations for this user or empty array if none
     */
    public static function get_registrations_by_user($userid = null, $activated = null) {
        global $DB;
        $results = array();
        $currentuserid = self::require_user();
        if (!isset($userid) && $currentuserid) {
            $userid = $currentuserid;
        }
        if (isset($userid)) {
            $sql = 'ownerid = ?'; // '".$userid."'";
            if (isset($activated)) {
                $sql .= ' and activated = '.($activated ? 1 : 0);
            }
            $results = $DB->get_records_select(self::REG_TABLENAME, $sql, array($userid), self::REG_ORDER);
            if (!$results) {
                $results = array();
            }
        }
        return $results;
    }

    /**
     * Convert an array of row data (one entry per field) into a CSV compatible string representing the row
     * @param array $row array of row data (one entry per field)
     * @return string CSV compatible row (includes the line breaks)
     */
    public static function make_csv_row($row) {
        $escapedrow = array();
        foreach ($row as $value) {
            $escapedrow[] = '"'.str_replace('"', '""', $value).'"';
        }
        $rowstring = implode(',', $escapedrow)."\r\n";
        return $rowstring;
    }

    /**
     * ADMIN ONLY
     * This is a method to get all the clickers for the clicker admin view
     * @param int $start [optional] start value for paging
     * @param int $max [optional] max value for paging
     * @param string $order [optional] the order by string
     * @param string $search [optional] search string for clickers
     * @param int $startdate [optional] starting registration date in unix seconds
     * @param int $enddate [optional] ending registration date in unix seconds
     * @return array the list of clicker registrations
     * @throws ClickerSecurityException if the current user is not allowed
     */
    public static function get_all_registrations($start = 0, $max = 0, $order = 'clickerid', $search = '', $startdate = null, $enddate = null) {
        global $DB;
        if (!self::is_admin()) {
            throw new ClickerSecurityException("Only admins can use this function");
        }
        if ($max <= 0) {
            $max = 10;
        }
        list($query, $params) = self::make_reg_filter_query($search, $startdate, $enddate);
        $results = $DB->get_records_select(self::REG_TABLENAME, $query, $params, $order, '*', $start, $max);
        if (!$results) {
            $results = array();
        } else {
            // insert user display names
            $userids = array();
            foreach ($results as $reg) {
                $userids[] = $reg->ownerid;
            }
            $userids = array_unique($userids);
            $users = self::get_users($userids);
            foreach ($results as $reg) {
                $name = 'UNKNOWN-'.$reg->ownerid;
                $email = 'UNKNOWN@unknown.com';
                if (array_key_exists($reg->ownerid, $users)) {
                    $name = $users[$reg->ownerid]->name;
                    $email = $users[$reg->ownerid]->email;
                }
                $reg->user_display_name = $name;
                $reg->user_email = $email;
            }
        }
        return $results;
    }

    /**
     * ADMIN ONLY
     * This is a method to get all the clickers for the clicker admin view
     * @param string $search [optional] search string for clickers
     * @param int $startdate [optional] starting registration date in unix seconds
     * @param int $enddate [optional] ending registration date in unix seconds
     * @return int the count of clicker registrations which were removed
     * @throws ClickerSecurityException if the current user is not allowed
     */
    public static function purge_registrations($search = '', $startdate = null, $enddate = null) {
        global $DB;
        if (!self::is_admin()) {
            throw new ClickerSecurityException("Only admins can use this function");
        }
        list($query, $params) = self::make_reg_filter_query($search, $startdate, $enddate);
        $count = $DB->count_records_select(self::REG_TABLENAME, $query, $params);
        // delete_records_select will throw an exception if the deletion fails
        $DB->delete_records_select(self::REG_TABLENAME, $query, $params);
        return $count;
    }

    /**
     * @param string $search [optional] search string for clickers
     * @param int $startdate [optional] starting registration date in unix seconds
     * @param int $enddate [optional] ending registration date in unix seconds
     * @return int the count of the total number of registered clickers
     */
    public static function count_all_registrations($search = '', $startdate = null, $enddate = null) {
        global $DB;
        list($query, $params) = self::make_reg_filter_query($search, $startdate, $enddate);
        return $DB->count_records_select(self::REG_TABLENAME, $query, $params);
    }

    /**
     * @param string $search [optional] search string for clickers
     * @param int $startdate [optional] starting registration date in unix seconds
     * @param int $enddate [optional] ending registration date in unix seconds
     * @return array of $query, $params
     */
    protected static function make_reg_filter_query($search, $startdate, $enddate) {
        global $DB;
        $query = '';
        $params = array();
        if (!empty($search)) {
            // build a search query
            $query .= $DB->sql_like('clickerid', '?', false, false, false); // clickerid like {search}%
            $params[] = $search . '%';
        }
        if (is_numeric($startdate) && is_numeric($enddate)) {
            $query .= (empty($query) ? '' : ' AND ') . '(timecreated >= ? AND timecreated <= ?)';
            $params[] = $startdate;
            $params[] = $enddate;

        } else if (is_numeric($startdate)) {
            $query .= (empty($query) ? '' : ' AND ') . 'timecreated >= ?';
            $params[] = $startdate;

        } else if (is_numeric($enddate)) {
            $query .= (empty($query) ? '' : ' AND ') . 'timecreated <= ?';
            $params[] = $enddate;

        }
        return array($query, $params);
    }

    /**
     * ADMIN ONLY
     * Removes the registration from the database
     *
     * @param int $regid id of the clicker registration
     * @return bool true if removed OR false if not found or not removed
     * @throws ClickerSecurityException
     */
    public static function remove_registration($regid) {
        global $DB;
        if (!self::is_admin()) {
            throw new ClickerSecurityException("Only admins can use this function");
        }
        if (isset($regid)) {
            if ($DB->delete_records(self::REG_TABLENAME, array('id' => $regid))) {
                return true;
            }
        }
        return false;
    }

    /**
     * Create a registration
     *
     * @param string $clickerid the clickerID (e.g. 11111111)
     * @param string $ownerid [optional] the userid OR current user if not set
     * @param boolean $localonly [optional] create this clicker in the local system only if true, otherwise sync to national system as well
     * @return stdClass the clicker_registration object
     * @throws ClickerRegisteredException
     */
    public static function create_clicker_registration($clickerid, $ownerid = null, $localonly = false) {
        $clickerid = self::validate_clicker_id($clickerid);
        $currentuserid = self::require_user();
        if (self::$allowremotesharing) {
            // when remote sharing is enabled we only look to see if this user has registered the remote
            $userid = $ownerid;
            if (!isset($ownerid)) {
                $userid = $currentuserid;
            }
        } else {
            // when remote sharing is disabled we fetch any that exist
            $userid = null;
        }
        $registration = self::get_registration_by_clicker_id($clickerid, $userid);
        // NOTE: we probably want to check the national system here to see if this is already registered
        if ($registration) {
            if ($registration->ownerid == $currentuserid) {
                // reactivate the clicker if needed
                if (!$registration->activated) {
                    $registration->activated = true;
                    self::save_registration($registration);
                }
            } else {
                $errmsg = 'Clicker '.$registration->clickerid.' already registered to '.$registration->ownerid;
                throw new ClickerRegisteredException($errmsg, $userid, $registration->clickerid, $registration->ownerid);
            }
        } else {
            $clickerregistration = new stdClass;
            $clickerregistration->clickerid = $clickerid;
            $clickerregistration->ownerid = $userid;
            $regid = self::save_registration($clickerregistration);
            $registration = self::get_registration_by_id($regid);
            if ($localonly) {
                // sync with national
                self::ws_sync_clicker($registration);
            }
        }
        return $registration;
    }

    /**
     * Make a registration active or inactive
     *
     * @param int $regid id of the clicker registration
     * @param boolean $activated true to enable, false to disable
     * @return stdClass the clicker_registration object
     * @throws InvalidArgumentException
     */
    public static function set_registration_active($regid, $activated) {
        if (!isset($regid)) {
            throw new InvalidArgumentException("reg_id must be set");
        }
        if (!isset($activated)) {
            throw new InvalidArgumentException("active must be set");
        }
        self::require_user();
        $registration = self::get_registration_by_id($regid);
        if (!$registration) {
            throw new InvalidArgumentException("Could not find registration with id ($regid)");
        }
        $registration->activated = $activated ? 1 : 0;
        self::save_registration($registration);
        return $registration;
    }

    /**
     * Saves the clicker registration data (create or update)
     * @param object $clickerregistration the registration data as an object
     * @return bool|int id of the saved registration
     * @throws ClickerSecurityException
     * @throws InvalidArgumentException if the registration is invalid (missing data or invalid data)
     */
    public static function save_registration(&$clickerregistration) {
        global $DB;
        if (!$clickerregistration || !isset($clickerregistration->clickerid)) {
            throw new InvalidArgumentException("clicker_registration cannot be empty and clickerid must be set");
        }
        $clickerregistration->clickerid = self::validate_clicker_id($clickerregistration->clickerid);
        $currentuserid = self::require_user();
        // set the owner to current if not set
        if (!isset($clickerregistration->ownerid)) {
            $clickerregistration->ownerid = $currentuserid;
        } else {
            // check for valid user id
            $user = self::get_users($clickerregistration->ownerid);
            if (! $user) {
                throw new InvalidArgumentException('User id ('.$clickerregistration->ownerid.') for registration is invalid');
            }
        }
        $clickerregistration->timemodified = time();
        if (!isset($clickerregistration->id)) {
            // new item to save (no perms check)
            $clickerregistration->timecreated = time();
            if (!$regid = $DB->insert_record(self::REG_TABLENAME, $clickerregistration, true)) {
                html_writer::tag('pre', s(var_dump($clickerregistration, true)), array('class' => 'notifytiny'));
                // Print_object($clickerregistration);.
                throw new moodle_exception('inserterror', self::BLOCK_NAME, '', null, null);
                // Print_error('inserterror', self::BLOCK_NAME);.
            }
        } else {
            // updating existing item
            if (self::can_write_registration($clickerregistration, $currentuserid)) {
                if (!$DB->update_record(self::REG_TABLENAME, $clickerregistration)) {
                    html_writer::tag('pre', s(var_dump($clickerregistration, true)), array('class' => 'notifytiny'));
                    // Print_object($clickerregistration);.
                    throw new moodle_exception('updateerror', self::BLOCK_NAME, '', null, null);
                    // Print_error('updateerror', self::BLOCK_NAME);.
                }
                $regid = $clickerregistration->id;
            } else {
                throw new ClickerSecurityException("Current user cannot update item ($clickerregistration->id) because they do not have permission");
            }
        }
        return $regid;
    }

    // COURSES METHODS

    /**
     * Get all the students for a course with their clicker registrations
     * @param int $courseid the course to get the students for
     * @param boolean $includeregs [optional]
     * @return array the list of user objects for the students in the course
     */
    public static function get_students_for_course_with_regs($courseid, $includeregs = true) {
        global $CFG, $DB;
        // get_users_by_capability - accesslib - moodle/grade:view
        // search_users - datalib
        if (class_exists('context_course')) {
            // for Moodle 2.2+
            $context = context_course::instance($courseid);
        } else {
             /* * @noinspection PhpDeprecationInspection */
            $context = context_module::instance($courseid);
        }
        $results = null;
        if ($CFG->version >= 2013051400) {
            // Moodle 2.5 included filter to only return active users.
            $results = get_enrolled_users($context, 'moodle/grade:view', 0, 'u.*', 'u.lastname', 0, 0, true);
        } else {
            $results = get_enrolled_users($context, 'moodle/grade:view', 0, 'u.*', 'u.lastname');
        }
        if (isset($results) && !empty($results)) {
            // get the registrations related to these students
            $userregs = array();
            if ($includeregs) {
                $query = 'activated = 1';
                if (count($results) > 500) {
                    // just return them all since the in query would be super slow anyway
                    $null = null;
                } else {
                    $query .= ' AND ownerid in (';
                    $first = true;
                    foreach ($results as $student) {
                        if ($first) {
                            $first = false;
                        } else {
                            $query .= ',';
                        }
                        $query .= '\''.$student->id.'\'';
                    }
                    $query .= ')';
                }
                $regs = $DB->get_records_select(self::REG_TABLENAME, $query);
                if ($regs) {
                    // now put these into a map
                    foreach ($regs as $reg) {
                        if (! array_key_exists($reg->ownerid, $userregs)) {
                            $userregs[$reg->ownerid] = array();
                        }
                        $userregs[$reg->ownerid][] = $reg;
                    }
                }
            }
            foreach ($results as $user) {
                // setup display name
                self::make_user_display_name($user);
                if ($includeregs) {
                    // add in registrations
                    $user->clicker_registered = false;
                    $user->clickers = array();
                    if (array_key_exists($user->id, $userregs)) {
                        $user->clicker_registered = true;
                        $user->clickers = $userregs[$user->id];
                    }
                }
            }
        } else {
            // NO matches
            $results = array();
        }
        return $results;
    }

    /**
     * Get the listing of all courses for an instructor
     * @param int $userid [optional] the unique user id for an instructor (default to current user)
     * @return array the list of courses (maybe be empty array)
     */
    public static function get_courses_for_instructor($userid = null) {
        // make this only get courses for this instructor
        // http://docs.moodle.org/en/Category:Capabilities - moodle/course:update
        if (! isset($userid)) {
            $userid = self::get_current_user_id();
        }
        if (class_exists('context_course')) {
            // NOTE: there is no maximum limit to this so self::$maxcoursestofetch has no effect
            // for Moodle 2.2+
            $results = enrol_get_users_courses($userid, true, array('fullname', 'summary', 'timecreated', 'visible'), null);
            foreach ($results as $id => $course) {
                $context = context_course::instance($id);
                if (!has_capability('moodle/course:update', $context, $userid)) {
                    unset($results[$id]);
                }
            }
        } else {
            // Moodle pre 2.2
            global $USER;
            if ($userid === $USER->id && isset($USER->access)) {
                $accessinfo = $USER->access;
            } else {
                $accessinfo = get_user_access_sitewide($userid);
            }
            $results = enrol_get_users_courses($userid, true, array('fullname', 'summary', 'timecreated', 'visible'), 'c.sortorder');
        }
        if (!$results) {
            $results = array();
        } else {
            // update the course titles
            foreach ($results as $course) {
                $course->title = $course->fullname;
                if (self::$enablecourseshortname && !empty($course->shortname)) {
                    $course->title = $course->fullname.' ('.$course->shortname.')';
                }
            }
        }
        return $results;
    }

    /**
     * Retrieve a single course by unique id
     * @param int $courseid the course
     * @return stdClass|bool the course object or false
     */
    public static function get_course($courseid) {
        global $DB;
        $course = $DB->get_record('course', array('id' => $courseid));
        // TESTING handling
        if (self::$testmode && !$course) {
            if ($courseid == '11111111') {
                $course = new stdClass();
                $course->id = $courseid;
                $course->fullname = 'testing: '.$courseid;
            }
        }
        if (!$course) {
            $course = false;
        }
        $course->title = $course->fullname;
        if (self::$enablecourseshortname && !empty($course->shortname)) {
            $course->title = $course->fullname.' ('.$course->shortname.')';
        }
        return $course;
    }

    /*  Not needed right now
        public static function get_course_grade_item($courseid, $gradeitemname) {
            if (! $courseid) {
                throw new InvalidArgumentException("course_id must be set");
            }
            if (! $gradeitemname) {
                throw new InvalidArgumentException("grade_item_name must be set");
            }
            $gradeitemfetched = false;
            $iclickercategory = grade_category::fetch(array(
                'courseid' => $courseid,
                'fullname' => self::GRADE_CATEGORY_NAME
                )
            );
            if ($iclickercategory) {
                $gradeitemfetched = grade_item::fetch(array(
                    'courseid' => $courseid,
                    'categoryid' => $iclicker_category->id,
                    'itemname' => $grade_item_name
                    )
                );
                if (! $gradeitemfetched) {
                    $gradeitemfetched = false;
                }
            }
            return $gradeitemfetched;
        }
    */

    private static function save_grade_item($gradeitem) {
        if (! $gradeitem) {
            throw new InvalidArgumentException("grade_item must be set");
        }
        if (! $gradeitem->courseid) {
            throw new InvalidArgumentException("grade_item->courseid must be set");
        }
        if (! $gradeitem->categoryid) {
            throw new InvalidArgumentException("grade_item->categoryid must be set");
        }
        if (! $gradeitem->name) {
            throw new InvalidArgumentException("grade_item->name must be set");
        }
        if (! isset($gradeitem->item_number)) {
            $gradeitem->item_number = 0;
        }

        // check for an existing item and update or create
        $gradeitemtosave = grade_item::fetch(array(
            'courseid' => $gradeitem->courseid,
            'itemmodule' => self::GRADE_ITEM_MODULE,
            // 'categoryid' => $gradeitem->categoryid,
            'itemname' => $gradeitem->name
            )
        );
        if (! $gradeitemtosave) {
            // create new one
            $gradeitemtosave = new grade_item();
            $gradeitemtosave->itemmodule = self::GRADE_ITEM_MODULE;
            $gradeitemtosave->courseid = $gradeitem->courseid;
            $gradeitemtosave->categoryid = $gradeitem->categoryid;
            $gradeitemtosave->iteminfo = $gradeitem->typename;
            // $gradeitemtosave->iteminfo = $gradeitem->name.' '.$gradeitem->type.' '.self::GRADE_CATEGORY_NAME;
            $gradeitemtosave->itemnumber = $gradeitem->item_number;
            // $gradeitemtosave->idnumber = $gradeitem->name;
            $gradeitemtosave->itemname = $gradeitem->name;
            $gradeitemtosave->itemtype = self::GRADE_ITEM_TYPE;
            // $gradeitemtosave->itemmodule = self::GRADE_ITEM_MODULE;
            if (isset($gradeitem->points_possible) && $gradeitem->points_possible > 0) {
                $gradeitemtosave->grademax = $gradeitem->points_possible;
            }
            $gradeitemtosave->insert(self::GRADE_LOCATION_STR);
        } else {
            // update
            if (isset($gradeitem->points_possible) && $gradeitem->points_possible > 0) {
                $gradeitemtosave->grademax = $gradeitem->points_possible;
            }
            $gradeitemtosave->categoryid = $gradeitem->categoryid;
            $gradeitemtosave->iteminfo = $gradeitem->typename;
            $gradeitemtosave->update(self::GRADE_LOCATION_STR);
        }
        $gradeitemid = $gradeitemtosave->id;
        $gradeitempp = $gradeitemtosave->grademax;

        // now save the related scores
        if (isset($gradeitem->scores) && !empty($gradeitem->scores)) {
            // get the existing scores
            $currentscores = array();
            $existinggrades = grade_grade::fetch_all(array(
                'itemid' => $gradeitemid
                )
            );
            if ($existinggrades) {
                foreach ($existinggrades as $grade) {
                    $currentscores[$grade->userid] = $grade;
                }
            }

            // run through the scores in the gradeitem and try to save them
            $errorscount = 0;
            $processedscores = array();
            foreach ($gradeitem->scores as $score) {
                $user = self::get_users($score->userid);
                if (! $user) {
                    $score->error = self::USER_DOES_NOT_EXIST_ERROR;
                    $processedscores[] = $score;
                    $errorscount++;
                    continue;
                }
                $userid = $user->id;
                // null/blank scores are not allowed
                if (! isset($score->score)) {
                    $score->error = 'NO_SCORE_ERROR';
                    $processedscores[] = $score;
                    $errorscount++;
                    continue;
                }
                if (! is_numeric($score->score)) {
                    $score->error = 'SCORE_INVALID';
                    $processedscores[] = $score;
                    $errorscount++;
                    continue;
                }
                $score->score = floatval($score->score);
                // Student Score should not be greater than the total points possible
                if ($score->score > $gradeitempp) {
                    $score->error = self::POINTS_POSSIBLE_UPDATE_ERRORS;
                    $processedscores[] = $score;
                    $errorscount++;
                    continue;
                }
                try {
                    $gradetosave = null;
                    if (isset($currentscores[$userid])) {
                        // existing score
                        $gradetosave = $currentscores[$userid];
                        // check against existing score
                        if ($score->score < $gradetosave->rawgrade) {
                            $score->error = self::SCORE_UPDATE_ERRORS;
                            $processedscores[] = $score;
                            $errorscount++;
                            continue;
                        }
                        $gradetosave->finalgrade = $score->score;
                        $gradetosave->rawgrade = $score->score;
                        $gradetosave->timemodified = time();
                         /* * @noinspection PhpUndefinedMethodInspection */
                        $gradetosave->update(self::GRADE_LOCATION_STR);
                    } else {
                        // new score
                        $gradetosave = new grade_grade();
                        $gradetosave->itemid = $gradeitemid;
                        $gradetosave->userid = $userid;
                        $gradetosave->finalgrade = $score->score;
                        $gradetosave->rawgrade = $score->score;
                        $gradetosave->rawgrademax = $gradeitempp;
                        $now = time();
                        $gradetosave->timecreated = $now;
                        $gradetosave->timemodified = $now;
                        $gradetosave->insert(self::GRADE_LOCATION_STR);
                    }
                     /* * @noinspection PhpUndefinedFieldInspection */
                    $gradetosave->userid = $score->userid;
                    $processedscores[] = $gradetosave;
                } catch (Exception $e) {
                    // General errors, caused while performing updates (Tag: generalerrors)
                    $score->error = self::GENERAL_ERRORS;
                    $processedscores[] = $score;
                    $errorscount++;
                }
            }
             /* * @noinspection PhpUndefinedFieldInspection */
            $gradeitemtosave->scores = $processedscores;
            // Put the errors in the item.
            if ($errorscount > 0) {
                $errors = array();
                foreach ($processedscores as $score) {
                    if (isset($score->error)) {
                        $errors[$score->userid] = $score->error;
                    }
                }
                 /* * @noinspection PhpUndefinedFieldInspection */
                $gradeitemtosave->errors = $errors;
            }
            $gradeitemtosave->force_regrading();
        }
        return $gradeitemtosave;
    }

    /**
     * Saves a gradebook (a set of grade items and scores related to a course),
     * also creates the categories based on the item type
     *
     * @param object $gradebook an object with at least course_id and items set
     * items should contain grade_items (courseid. categoryid, name, scores)
     * scores should contain grade_grade (userid, score)
     * @return stdClass the saved gradebook with all items and scores in the same structure,
     * errors are recorded as grade_item->errors and score->error
     * @throws InvalidArgumentException
     */
    public static function save_gradebook($gradebook) {
        global $CFG;
        if (! $gradebook) {
            throw new InvalidArgumentException("gradebook must be set");
        }
        if (! isset($gradebook->course_id)) {
            throw new InvalidArgumentException("gradebook->course_id must be set");
        }
        if (! isset($gradebook->items) || empty($gradebook->items)) {
            throw new InvalidArgumentException("gradebook->items must be set and include items");
        }
        $gbsaved = new stdClass();
        $gbsaved->items = array();
        $gbsaved->course_id = $gradebook->course_id;
        $course = self::get_course($gradebook->course_id);
        if (! $course) {
            throw new InvalidArgumentException("No course found with course_id ($gradebook->course_id)");
        }
        $gbsaved->course = $course;

        // Extra permissions check on gradebook manage/update.
        $userid = self::require_user();
        if (class_exists('context_course')) {
            // for Moodle 2.2+
            $context = context_course::instance($course->id);
        } else {
             /* * @noinspection PhpDeprecationInspection */
            $context = context_module::instance($course->id);
        }
        if (!$context || !has_capability('moodle/grade:manage', $context, $userid)) {
            $excp = "User ($userid) cannot manage the gradebook in course id=$course->id (".var_export($course, true)."), context: ".var_export($context, true);
            throw new InvalidArgumentException($excp);
        }

        // attempt to get the default iclicker category first
        $defaulticlickercategory = grade_category::fetch(array(
            'courseid' => $gradebook->course_id,
            'fullname' => self::GRADE_CATEGORY_NAME
            )
        );
        $defaulticlickercategoryid = $defaulticlickercategory ? $defaulticlickercategory->id : null;
        // echo "\n\nGRADEBOOK: ".var_export($gradebook);
        // iterate through and save grade items by calling other method
        if (! empty($gradebook->items)) {
            $saveditems = array();
            $number = 0;
            foreach ($gradebook->items as $gradeitem) {
                // check for this category
                $itemcategoryname = self::GRADE_CATEGORY_NAME;
                if (! empty($gradeitem->type) && self::GRADE_CATEGORY_NAME != $gradeitem->type) {
                    $itemcategoryname = $gradeitem->type;
                    $itemcategory = grade_category::fetch(array(
                        'courseid' => $gradebook->course_id,
                        'fullname' => $itemcategoryname
                        )
                    );
                    if (! $itemcategory) {
                        // create the category
                        $params = array(
                            'courseid' => $gradebook->course_id,
                            'fullname' => $itemcategoryname,
                        );
                        $gradecategory = new grade_category($params, false);
                        // Use default aggregation type.
                        $gradecategory->aggregation = $CFG->grade_aggregation;
                        $gradecategory->insert(self::GRADE_LOCATION_STR);
                        $itemcategoryid = $gradecategory->id;
                    } else {
                        $itemcategoryid = $itemcategory->id;
                    }
                } else {
                    // use default
                    if (! $defaulticlickercategoryid) {
                        // create the category
                        $params = array(
                            'courseid' => $gradebook->course_id,
                            'fullname' => self::GRADE_CATEGORY_NAME,
                        );
                        $gradecategory = new grade_category($params, false);
                        // Use default aggregation type.
                        $gradecategory->aggregation = $CFG->grade_aggregation;
                        $gradecategory->insert(self::GRADE_LOCATION_STR);
                        $defaulticlickercategoryid = $gradecategory->id;
                    }
                    $itemcategoryid = $defaulticlickercategoryid;
                }
                $gradeitem->categoryid = $itemcategoryid;
                $gradeitem->typename = $itemcategoryname;
                $gradeitem->courseid = $gradebook->course_id;
                $gradeitem->item_number = $number;
                $savedgradeitem = self::save_grade_item($gradeitem);
                $saveditems[] = $savedgradeitem;
                $number++;
            }
            $gbsaved->items = $saveditems;
        }
        $gbsaved->default_category_id = $defaulticlickercategoryid;
        // echo "\n\nRESULT: ".var_export($gbsaved);
        return $gbsaved;
    }

    // DATA ENCODING METHODS

    /**
     * Encodes a clicker registration into XML
     *
     * @param object $clickerregistration fields(ownerid, clickerid)
     * @return string the XML
     * @throws InvalidArgumentException if the registration is invalid
     */
    public static function encode_registration($clickerregistration) {
        if (! $clickerregistration) {
            throw new InvalidArgumentException("clicker_registration must be set");
        }
        $userid = $clickerregistration->ownerid;
        $user = self::get_users($userid);
        if (! $user) {
            throw new InvalidArgumentException("Invalid user id ($userid) for clicker reg ($clickerregistration->clickerid)");
        }
        if (!isset($clickerregistration->activated)) {
            $clickerregistration->activated = true;
        }
        $encoded = '<Register>'.PHP_EOL;
        $encoded .= '  <S DisplayName="';
        $encoded .= self::encode_for_xml($user->name);
        $encoded .= '" FirstName="';
        $encoded .= self::encode_for_xml($user->firstname);
        $encoded .= '" LastName="';
        $encoded .= self::encode_for_xml($user->lastname);
        $encoded .= '" StudentID="';
        $encoded .= self::encode_for_xml(strtoupper($user->username));
        $encoded .= '" Email="';
        $encoded .= self::encode_for_xml($user->email);
        $encoded .= '" URL="';
        $encoded .= self::encode_for_xml(self::$domainurl);
        $encoded .= '" ClickerID="';
        $encoded .= strtoupper($clickerregistration->clickerid);
        $encoded .= '" Enabled="';
        $encoded .= $clickerregistration->activated ? 'True' : 'False';
        $encoded .= '"></S>'.PHP_EOL;
        // close out
        $encoded .= '</Register>'.PHP_EOL;
        return $encoded;
    }

    /**
     * Creates XML encoding of the result of a clicker registration
     *
     * @param object $registrations [optional] all regs for the registering user
     * @param boolean $status true if success, false if failure
     * @param string $message the message to send along (typically failure message)
     * @return string the XML
     * @throws InvalidArgumentException if the data is invalid
     */
    public static function encode_registration_result($registrations, $status, $message) {
        if (! $registrations) {
            throw new InvalidArgumentException("registrations must be set");
        }
        if (! isset($status)) {
            throw new InvalidArgumentException("status must be set");
        }
        /* SAMPLE
        1) When clicker is already registered to some one else - the same
        message should be returned that is displayed in the plug-in in xml
        format
        <RetStatus Status="False" Message=""/>

        2) When clicker is already registered to the same user - the same
        message should be returned that is displayed in the plug-in in xml
        format.
        <RetStatus Status="False" Message=""/>

        3) When studentid is not found in the CMS
        <RetStatus Status="False" Message="Student not found in the CMS"/>

        4) Successful registration -
        <RetStatus Status="True" Message="..."/>
         */
        $encoded = '<RetStatus Status="'.($status ? 'True' : 'False').'" Message="'.self::encode_for_xml($message).'" />';
        return $encoded;
    }

    /**
     * Encode a set of courses which a user is an instructor for into XML
     *
     * @param int $instructorid unique user id
     * @return string the XML
     * @throws InvalidArgumentException if the id is invalid
     * @throws ClickerSecurityException
     */
    public static function encode_courses($instructorid) {
        if (! isset($instructorid)) {
            throw new InvalidArgumentException("instructor_id must be set");
        }
        $instructor = self::get_users($instructorid);
        if (! $instructor) {
            throw new InvalidArgumentException("Invalid instructor user id ($instructorid)");
        }
        $courses = self::get_courses_for_instructor($instructorid);
        if (! $courses) {
            throw new ClickerSecurityException("No courses found, only instructors can access instructor courses listings");
        }
        $encoded = '<coursemembership username="';
        $encoded .= self::encode_for_xml($instructor->username);
        $encoded .= '">'.PHP_EOL;
        // loop through courses
        foreach ($courses as $course) {
            $coursename = $course->fullname;
            if (self::$enablecourseshortname && !empty($course->shortname)) {
                $coursename .= '-'.$course->shortname;
            }
            $encoded .= '  <course id="'.$course->id.'" name="'.self::encode_for_xml($coursename)
                    .'" created="'.$course->timecreated
                    .'" published="'.($course->visible ? 'True' : 'False')
                    .'" usertype="I" />'.PHP_EOL;
        }
        // close out
        $encoded .= '</coursemembership>'.PHP_EOL;
        return $encoded;
    }

    /**
     * Encode a set of enrollments for a course into XML
     *
     * @param int $courseid unique id for a course
     * @return string the XML
     * @throws InvalidArgumentException if the id is invalid
     */
    public static function encode_enrollments($courseid) {
        if (! isset($courseid)) {
            throw new InvalidArgumentException("course_id must be set");
        }
        $course = self::get_course($courseid);
        if (! $course) {
            throw new InvalidArgumentException("No course found with course_id ($courseid)");
        }
        $students = self::get_students_for_course_with_regs($courseid);
        $encoded = '<courseenrollment courseid="'.$course->id.'">'.PHP_EOL;
        // the students may be an empty set
        if ($students) {
            // loop through students
            foreach ($students as $student) {
                // get the clicker data out first if there is any
                $cidsdates = self::make_clicker_ids_and_dates($student->clickers);
                // now make the actual user data line
                $encoded .= '  <user id="'.$student->id.'" usertype="S" firstname="';
                $encoded .= self::encode_for_xml($student->firstname ? $student->firstname : '');
                $encoded .= '" lastname="';
                $encoded .= self::encode_for_xml($student->lastname ? $student->lastname : '');
                $encoded .= '" emailid="';
                $encoded .= self::encode_for_xml($student->email ? $student->email : '');
                $encoded .= '" uniqueid="';
                $encoded .= self::encode_for_xml($student->username);
                $encoded .= '" clickerid="';
                $encoded .= self::encode_for_xml( $cidsdates['clickerid'] );
                $encoded .= '" whenadded="';
                $encoded .= self::encode_for_xml( $cidsdates['whenadded'] );
                $encoded .= '" />'.PHP_EOL;
            }
        }
        // close out
        $encoded .= '</courseenrollment>'.PHP_EOL;
        return $encoded;
    }

    /**
     * @static
     * @param array $clickerregs array of clicker registrations
     * @return array an array with 2 strings under these keys ('clickerid', 'whenadded')
     */
    public static function make_clicker_ids_and_dates($clickerregs) {
        $clickerids = '';
        $clickeraddeddates = '';
        if ($clickerregs && !empty($clickerregs)) {
            $count = 0;
            foreach ($clickerregs as $reg) {
                if ($count > 0) {
                    $clickerids .= ',';
                    $clickeraddeddates .= ',';
                }
                $clickerids .= $reg->clickerid;
                $clickerdate = date('M/d/Y', $reg->timecreated);
                $clickeraddeddates .= $clickerdate;
                $count++;
                if (! self::$disablealternateid) {
                    // add in the alternate clicker id if needed
                    $alternateid = self::translate_clicker_id($reg->clickerid);
                    if ($alternateid != null) {
                        $clickerids .= ','.$alternateid;
                        $clickeraddeddates .= ','.$clickerdate;
                        $count++;
                    }
                }
            }
        }
        return array('clickerid' => $clickerids, 'whenadded' => $clickeraddeddates);
    }

    /**
     * Encodes the results of a gradebook save into XML
     *
     * @param object $gradebookresult the result from gradebook_save
     * @return string the XML
     * @throws InvalidArgumentException if the registration is invalid
     */
    public static function encode_gradebook_results($gradebookresult) {
        if (! isset($gradebookresult->course)) {
            throw new InvalidArgumentException("course must be set");
        }
        // $course = $gradebookresult->course;
        $courseid = $gradebookresult->course->id;
        // check for any errors
        $haserrors = false;
        foreach ($gradebookresult->items as $item) {
            if (isset($item->errors) && !empty($item->errors)) {
                $haserrors = true;
                break;
            }
        }
        /* SAMPLE
        <errors courseid="BFW61">
          <Userdoesnotexisterrors>
            <user id="XXXX" />
          </Userdoesnotexisterrors>
          <Scoreupdateerrors>
            <user id="2222">
              <lineitem name="Decsample" pointspossible="0" type="Text" score="9" />
            </user>
          </Scoreupdateerrors>
          <PointsPossibleupdateerrors>
            <user id="33333">
              <lineitem name="CMSIntTEST01" pointspossible="50" type="iclicker polling scores" score="70" />
            </user>
          </PointsPossibleupdateerrors>
          <Scoreupdateerrors>
            <user id="444444">
              <lineitem name="Mac-integrate-2" pointspossible="31" type="092509Mac" score="13"/>
            </user>
          </Scoreupdateerrors>
          <Generalerrors>
            <user id="5555" error="CODE">
              <lineitem name="itemName" pointspossible="35" score="XX" error="CODE" />
            </user>
          </Generalerrors>
        </errors>
         */
        $output = null;
        if ($haserrors) {
            $lineitems = self::make_lineitems($gradebookresult->items);
            $invaliduserids = array();

            $encoded = '<errors courseId="'.$courseid.'">'.PHP_EOL;
            // loop through items and errors and generate errors xml blocks
            $erroritems = array();
            foreach ($gradebookresult->items as $item) {
                if (isset($item->errors) && !empty($item->errors)) {
                    foreach ($item->scores as $score) {
                        if (isset($score->error)) {
                            $lineitem = $lineitems[$item->id];
                            if (self::USER_DOES_NOT_EXIST_ERROR == $score->error) {
                                $key = self::USER_DOES_NOT_EXIST_ERROR;
                                if (! array_key_exists($score->userid, $invaliduserids)) {
                                    // only if the invalid user is not already listed in the errors
                                    if (!isset($erroritems[$key])) {
                                        $erroritems[$key] = '';
                                    }
                                    $erroritems[$key] .= '    <user id="'.$score->userid.'" />'.PHP_EOL;
                                    $invaliduserids[$score->userid] = $score->userid;
                                }
                            } else if (self::POINTS_POSSIBLE_UPDATE_ERRORS == $score->error) {
                                $key = self::POINTS_POSSIBLE_UPDATE_ERRORS;
                                $li = str_replace(self::SCORE_KEY, $score->score, $lineitem);
                                if (!isset($erroritems[$key])) {
                                    $erroritems[$key] = '';
                                }
                                $erroritems[$key] .= '    <user id="'.$score->userid.'">'.PHP_EOL.'      '.$li.PHP_EOL.'    </user>'.PHP_EOL;
                            } else if (self::SCORE_UPDATE_ERRORS == $score->error) {
                                $key = self::SCORE_UPDATE_ERRORS;
                                $li = str_replace(self::SCORE_KEY, $score->score, $lineitem);
                                if (!isset($erroritems[$key])) {
                                    $erroritems[$key] = '';
                                }
                                $erroritems[$key] .= '    <user id="'.$score->userid.'">'.PHP_EOL.'      '.$li.PHP_EOL.'    </user>'.PHP_EOL;
                            } else {
                                // general error
                                $key = self::GENERAL_ERRORS;
                                $li = str_replace(self::SCORE_KEY, $score->score, $lineitem);
                                if (!isset($erroritems[$key])) {
                                    $erroritems[$key] = '';
                                }
                                $erroritems[$key] .= '    <user id="'.$score->userid.'" error="'.$score->error.'">'.PHP_EOL.
                                                      '      <error type="'.$score->error.'" />'.PHP_EOL.
                                                      '      '.$li.PHP_EOL.
                                                      '    </user>'.PHP_EOL;
                            }
                        }
                    }
                }
            }
            // loop through error items and dump to the output
            if (array_key_exists(self::USER_DOES_NOT_EXIST_ERROR, $erroritems)) {
                $encoded .= '  <Userdoesnotexisterrors>'.PHP_EOL.$erroritems[self::USER_DOES_NOT_EXIST_ERROR].'  </Userdoesnotexisterrors>'.PHP_EOL;
            }
            if (array_key_exists(self::POINTS_POSSIBLE_UPDATE_ERRORS, $erroritems)) {
                $encoded .= '  <PointsPossibleupdateerrors>'.PHP_EOL.$erroritems[self::POINTS_POSSIBLE_UPDATE_ERRORS].'  </PointsPossibleupdateerrors>'.PHP_EOL;
            }
            if (array_key_exists(self::SCORE_UPDATE_ERRORS, $erroritems)) {
                $encoded .= '  <Scoreupdateerrors>'.PHP_EOL.$erroritems[self::SCORE_UPDATE_ERRORS].'  </Scoreupdateerrors>'.PHP_EOL;
            }
            if (array_key_exists(self::GENERAL_ERRORS, $erroritems)) {
                $encoded .= '  <Generalerrors>'.PHP_EOL.$erroritems[self::GENERAL_ERRORS].'  </Generalerrors>'.PHP_EOL;
            }
            // close out
            $encoded .= '</errors>'.PHP_EOL;
            $output = $encoded;
        }
        return $output;
    }

    private static function make_lineitems($items) {
        $lineitems = array();
        foreach ($items as $item) {
            $li = '<lineitem name="'.self::encode_for_xml($item->itemname).'" pointspossible="'.$item->grademax.'" type="'.$item->iteminfo.'" score="'.self::SCORE_KEY.'"/>';
            $lineitems[$item->id] = $li;
        }
        return $lineitems;
    }

    /**
     * This will handle the initial parsing of an XML string into a DOM document
     *
     * @param string $xml the xml string
     * @return DOMDocument object
     * @throws InvalidArgumentException if the xml is not set
     * @throws DOMException if the xml fails to parse
     * @throws Exception on xml load failure
     */
    private static function parse_xml_to_doc($xml) {
        if (! $xml) {
            throw new InvalidArgumentException("xml must be set");
        }
        // read the xml (try to anyway)
        set_error_handler('iclicker_handlexmlerror');
        $doc = new DOMDocument();
        $doc->preserveWhiteSpace = false;
        if (! $doc->loadXML($xml, LIBXML_NOWARNING) ) {
            throw new Exception("XML read and parse failure: $xml");
        }
        $doc->normalizeDocument();
        restore_error_handler();
        return $doc;
    }

    /**
     * Translate incoming XML into a clicker registration,
     * will figure out the user and get necessary data
     *
     * @param string $xml the xml
     * @return stdClass the clicker_registration object
     * @throws InvalidArgumentException if the xml cannot be parsed
     * @throws Exception
     */
    public static function decode_registration($xml) {
        /*
        <Register>
          <S DisplayName="DisplayName-azeckoski-123456" FirstName="First" LastName="Lastazeckoski-123456"
            StudentID="123456" Email="azeckoski-123456@email.com" URL="http://sakaiproject.org" ClickerID="11111111"></S>
        </Register>
         */
        $doc = self::parse_xml_to_doc($xml);
        $clickerreg = new stdClass();
        try {
            $users = $doc->getElementsByTagName("S");
            if ($users->length <= 0) {
                throw new InvalidArgumentException("Invalid XML, no S element");
            }
            $usernode = $users->item(0);
            if ($usernode->nodeType == XML_ELEMENT_NODE) {
                 /* * @noinspection PhpUndefinedMethodInspection */
                $clickerid = $usernode->getAttribute("ClickerID");
                if (! $clickerid) {
                    throw new InvalidArgumentException("Invalid XML for registration, no id in the ClickerID element (Cannot process)");
                }
                 /* * @noinspection PhpUndefinedMethodInspection */
                $userid = $usernode->getAttribute("StudentID"); // this is the username
                if (! $userid) {
                    throw new InvalidArgumentException("Invalid XML for registration, no id in the StudentID element (Cannot process)");
                }
                $clickerreg->clickerid = $clickerid;
                $user = self::get_users($userid);
                if (! $user) {
                    throw new InvalidArgumentException("Invalid username for student ($userid), could not find user (Cannot process)");
                }
                $clickerreg->user_username = $user->username;
                $clickerreg->ownerid = $user->id;
                 /* * @noinspection PhpUndefinedMethodInspection */
                $clickerreg->user_display_name = $usernode->getAttribute("DisplayName");
            } else {
                throw new InvalidArgumentException("Invalid user node in XML: $usernode");
            }
        } catch (Exception $e) {
            throw new Exception("XML DOM parsing failure: $e :: $xml");
        }
        return $clickerreg;
    }

    /**
     * Decodes XML into a gradebook object
     *
     * @param string $xml the xml
     * @return stdClass the gradebook object
     * @throws InvalidArgumentException if the xml cannot be parsed or the data is invalid
     * @throws Exception
     */
    public static function decode_gradebook($xml) {
        /*
        <coursegradebook courseid="BFW61">
          <user id="id01" usertype="S">
            <lineitem name="06/02/2009" pointspossible="50" type="iclicker polling scores" score="0"/>
          </user>
          <user id="id02" usertype="S">
            <lineitem name="06/02/2009" pointspossible="50" type="iclicker polling scores" score="0"/>
          </user>
        </coursegradebook>
        */
        $doc = self::parse_xml_to_doc($xml);
        $gradebook = new stdClass();
        $gradebook->students = array();
        $gradebook->items = array();
        try {
            // get the course id from the root attribute
            $courseid = $doc->documentElement->getAttribute("courseid");
            if (! $courseid) {
                throw new InvalidArgumentException("Invalid XML, no courseid in the root xml element");
            }
            $users = $doc->getElementsByTagName("user");
            if ($users->length <= 0) {
                throw new InvalidArgumentException("Invalid XML, no user elements found");
            }
            $gradebook->course_id = $courseid;
            foreach ($users as $usernode) {
                if ($usernode->nodeType == XML_ELEMENT_NODE) {
                     /* * @noinspection PhpUndefinedMethodInspection */
                    $usertype = $usernode->getAttribute("usertype");
                    if (strcasecmp('s', $usertype) != 0) {
                        continue; // Skip this one.
                    }
                    // Valid user to process.
                    /* * @noinspection PhpUndefinedMethodInspection */
                    $userid = $usernode->getAttribute("id"); // this is the user id (not username)
                    if (! $userid) {
                        $debugmsg = "WARN: Gradebook import failure for course ($courseid), Invalid XML for user, no id in the user element (skipping this entry): ";
                        $debugmsg .= var_export($usernode);
                        debugging($debugmsg);
                        continue;
                    }
                    /* check the username when saving
                    $user = self::get_user_by_username($username);
                    if (! $user) {
                        throw new InvalidArgumentException("Invalid username for student ($username), could not find user (Cannot process)");
                    }
                    $userid = $user->id;
                    */
                    $gradebook->students[$userid] = $userid;
                     /* * @noinspection PhpUndefinedMethodInspection */
                    $lineitems = $usernode->getElementsByTagName("lineitem");
                    foreach ($lineitems as $lineitem) {
                         /* * @noinspection PhpUndefinedMethodInspection */
                        $liname = $lineitem->getAttribute("name");
                        if (! $liname) {
                            throw new InvalidArgumentException("Invalid XML, no name in the lineitem xml element: $lineitem");
                        }
                        $gradeitem = null;
                        if (! isset($gradebook->items[$liname])) {
                            // only add lineitem from the first item
                            /* * @noinspection PhpUndefinedMethodInspection */
                            $litype = $lineitem->getAttribute("type");
                            $lipp = 100.0;
                            /* * @noinspection PhpUndefinedMethodInspection */
                            $lipptext = $lineitem->getAttribute("pointspossible");
                            if (isset($lipptext) && $lipptext != '') {
                                if (!is_numeric($lipptext)) {
                                    $errmsg = "WARN: Gradebook import failure for course ($courseid) and user ($userid),";
                                    $errmsg .= " Invalid points possible ($lipptext), using default of $lipp";
                                    debugging($errmsg);
                                } else {
                                    $lipp = floatval($lipptext);
                                }
                            }
                            $gradeitem = new stdClass();
                            $gradeitem->name = $liname;
                            $gradeitem->points_possible = $lipp;
                            $gradeitem->type = trim($litype);
                            $gradeitem->scores = array();
                            $gradebook->items[$liname] = $gradeitem;
                        } else {
                            $gradeitem = $gradebook->items[$liname];
                        }
                        /* * @noinspection PhpUndefinedMethodInspection */
                        $liscore = $lineitem->getAttribute("score");
                        if (! isset($liscore) || '' == $liscore) {
                            $errmsg = "WARN: Gradebook import failure for course ($courseid) and user ($userid),";
                            $errmsg .= " Invalid score ($liscore), skipping this entry: ".var_export($lineitem);
                            debugging($errmsg);
                            continue;
                        }
                        // add in the score
                        $score = new stdClass();
                        $score->item_name = $gradeitem->name;
                        $score->userid = $userid;
                        $score->score = $liscore;
                        $gradeitem->scores[] = $score;
                    }
                } else {
                    throw new InvalidArgumentException("Invalid user node in XML: $usernode");
                }
            }
        } catch (Exception $e) {
            throw new Exception("XML DOM parsing failure: $e :: $xml");
        }
        return $gradebook;
    }

    /**
     * Decodes the webservices xml into an array of clicker registration objects
     *
     * @param string $xml the xml from an iclicker webservice
     * @return array (clicker_registration object)
     * @throws InvalidArgumentException if the xml cannot be parsed or the data is invalid
     * @throws Exception
     */
    public static function decode_ws_xml($xml) {
        /*
        <StudentRoster>
            <S StudentID="student01" FirstName="student01" LastName="student01" URL="https://www.iclicker.com/" CourseName="">
                <Registration ClickerId="12CE32EE" WhenAdded="2009-01-27" Enabled="True" />
            </S>
        </StudentRoster>
        */
        $doc = self::parse_xml_to_doc($xml);
        $regs = array();
        try {
            $users = $doc->getElementsByTagName("S");
            if ($users->length > 0) {
                foreach ($users as $usernode) {
                    if ($usernode->nodeType == XML_ELEMENT_NODE) {
                         /* * @noinspection PhpUndefinedMethodInspection */
                        $studentid = $usernode->getAttribute("StudentID"); // this is the user eid
                        if (! isset($studentid) || '' == $studentid) {
                            throw new InvalidArgumentException("Invalid XML for registration, no id in the StudentID element (Cannot process)");
                        }
                        $studentid = strtolower($studentid); // username
                        $user = self::get_user_by_username($studentid);
                        if (! $user) {
                            // log.warn("Cannot identify user (id="+studentId+") in the national webservices feed, skipping this user");
                            continue;
                        }
                        $userid = $user->id;
                        /* * @noinspection PhpUndefinedMethodInspection */
                        $regnodes = $usernode->getElementsByTagName("Registration");
                        if ($regnodes->length > 0) {
                            foreach ($regnodes as $regnode) {
                                if ($regnode->nodeType == XML_ELEMENT_NODE) {
                                    /* * @noinspection PhpUndefinedMethodInspection */
                                    $clickerid = $regnode->getAttribute("ClickerId");
                                    if (! $clickerid) {
                                        // log.warn("Missing clickerId in webservices registration XML line, skipping this registration for user: $userid");
                                        continue;
                                    }
                                    /* * @noinspection PhpUndefinedMethodInspection */
                                    $whenadded = $regnode->getAttribute("WhenAdded"); // "yyyy-MM-dd"
                                    $datecreated = time();
                                    if (isset($whenadded)) {
                                        $time = strtotime($whenadded);
                                        if ($time) {
                                            $datecreated = $time;
                                        }
                                    }
                                    /* * @noinspection PhpUndefinedMethodInspection */
                                    $enabled = $regnode->getAttribute("Enabled");
                                    $activated = true;
                                    if (isset($enabled)) {
                                        $activated = (boolean) $enabled;
                                    }
                                    $clickerreg = new stdClass();
                                    $clickerreg->clickerid = $clickerid;
                                    $clickerreg->ownerid = $userid;
                                    $clickerreg->user_username = $studentid;
                                    $clickerreg->timecreated = $datecreated;
                                    $clickerreg->date_created = $datecreated;
                                    $clickerreg->activated = $activated;
                                    $regs[] = $clickerreg;
                                } else {
                                    // only skipping invalid ones
                                    // log.warn("Invalid registration node in XML (skipping this one): $regnode");
                                    $null = null;
                                }
                            }
                        }
                    } else {
                        // only skipping invalid ones
                        // throw new InvalidArgumentException("Invalid user node in XML: $usernode");
                        $null = null;
                    }
                }
            }
        } catch (Exception $e) {
            throw new Exception("XML DOM parsing failure: $e :: $xml");
        }
        return $regs;
    }

    // GO WEBSERVICES

    /**
     * Verify a clicker id with the GO webservices server,
     * Returns true on success OR ClickerIdInvalidException on failure
     *
     * @static
     * @param string $clickergold 12 char clicker go id
     * @param string $studentlastname
     * @return bool true if the clicker id is valid and linked to the provided user lastname (or throws exception)
     * @throws ClickerIdInvalidException (GO_LASTNAME) if the lastname does not match OR GO_NO_MATCH if it does not match
     * @throws ClickerWebservicesException if the format does not match or a failure occurs
     * @throws InvalidArgumentException if the params are not set
     */
    public static function ws_go_verify_clickerid($clickergold, $studentlastname) {
        if (!isset($clickergold) || !isset($studentlastname)) {
            throw new InvalidArgumentException("clickerGOId=$clickergold and studentLastName=$studentlastname must both be set");
        }
        $wsoperation = 'GetRegisteredForClickerMAC';
        $arguments = array(
                'pVarClickerID' => base64_encode($clickergold),
        );
        // echo 'args: '.var_export(array($wsoperation, $arguments),true).''.PHP_EOL;
        $result = self::ws_go_soap_call($wsoperation, $arguments);
        // echo 'result: ' . var_export($result, true) . '' . PHP_EOL;
        $xml = $result['GetRegisteredForClickerMACResult'];
        if (empty($xml)) {
            // no registration matches
            throw new ClickerIdInvalidException("No match found on the server for clicker ($clickergold)", ClickerIdInvalidException::GO_NO_MATCH, $clickergold);
        } else {
            // <StudentEnrol><S StudentId="testgoqait99" FirstName="testgoqait99" LastName="testgoqait99" MiddleName="" WebClickerId="C570BF0C2154"/></StudentEnrol>
            $xml = base64_decode($xml);
            // echo 'XML: ' . var_export($xml, true) . '' . PHP_EOL;
            $enrollment = @simplexml_load_string($xml);
            // echo 'enrollment: ' . var_export($enrollment, true) . '' . PHP_EOL;
            if ($enrollment && isset($enrollment->S) && isset($enrollment->S->attributes()->LastName)) {
                $lastname = $enrollment->S->attributes()->LastName;
                $verified = strcasecmp($studentlastname, $lastname) == 0;
                // echo "verified: $studentlastname == $lastname: " . var_export($verified, true) . '' . PHP_EOL;
                if (!$verified) {
                    // should we log a warning here? -AZ
                    $errmsg = "Lastname ($studentlastname) does not match with registered lastname ($lastname) for clicker ($clickergold)";
                    throw new ClickerIdInvalidException($errmsg, ClickerIdInvalidException::GO_LASTNAME, $clickergold);
                }
            } else {
                // structure did not properly match
                $msg = "i>clicker Webservices return structure does not match expected format (please contact support): $xml";
                debugging($msg);
                throw new ClickerWebservicesException($msg);
            }
        }
        return $verified;
    }

    /**
     * Handles the soap call to the GO webservices server
     *
     * @static
     * @param string $wsoperation the operation to perform (e.g. 'StudentsReport')
     * @param array $wsarguments the SOAP arguments to send
     * @return array the results of the SOAP call
     * @throws ClickerWebservicesException if the call fails
     */
    private static function ws_go_soap_call($wsoperation, $wsarguments) {
        $client = new SoapClient('https://www.iclickergo.com/webservice/webvoting.asmx?WSDL', array(
                        'soap_version' => SOAP_1_2,
                        'encoding' => 'UTF-8',
                        'trace' => 1,
                        'exceptions' => 1,
                )
        );
        try {
            $result = $client->__soapCall($wsoperation, array('parameters' => $wsarguments));
            if ($result instanceof SoapFault) {
                $msg = 'SoapFault occurred: ' . var_export($result, true);
                debugging($msg);
                throw new Exception($msg);
            }
        } catch (Exception $e) {
            $msg = 'Failure in the i>clicker webservices: ' . var_export($client, true);
            debugging($msg);
            throw new ClickerWebservicesException($msg);
        }
        // convert result to an array
        $result = json_decode(json_encode($result), true);
        // echo '<pre>SOAP: '.var_export($client,true).'</pre>'.PHP_EOL;
        unset($client);
        return $result;
    }


    // NATIONAL WEBSERVICES

    /**
     * Syncs this clicker with the national services clicker (ensure that this is saved to national)
     *
     * @static
     * @param stdClass $clickerregistration
     * @return array|bool array('errors') with errors if any occurred, false if national ws is disabled
     */
    public static function ws_sync_clicker($clickerregistration) {
        $results = array('errors' => array());
        if (self::$usenationalwebservices) {
            try {
                $regs = self::ws_save_clicker($clickerregistration);
                if ($regs && count($regs) > 1) {
                    $crkey = self::make_reg_key($clickerregistration);
                    $syncregs = array();
                    foreach ($regs as $reg) {
                        $id = self::make_reg_key($reg);
                        $syncregs[$id] = $reg;
                    }
                    unset($syncregs[$crkey]);
                    $regs = array_values($syncregs);
                    // now we save all the registrations from national (may already exist)
                    foreach ($regs as $reg) {
                        try {
                            self::ws_save_clicker($reg);
                        } catch (Exception $e) {
                            // this ok, we will carry on
                            $results['errors'][] = $reg;
                        }
                    }
                }
            } catch (Exception $e) {
                // failed to sync with national but we do not fail, only send notification and log
                $msg = "Failure while syncing i>clicker registration ($clickerregistration): $e";
                $results['exception'] = $e;
                self::send_notifications($msg, $e);
                // log.warn(msg);
            }
        } else {
            $results = false;
        }
        return $results;
    }

    /**
     * Syncs all current clickers with the national services clickers for this site
     *
     * @return array results array('errors') with errors if any occurred, false if national ws is disabled
     * @throws Exception
     */
    public static function ws_sync_all() {
        $results = array('errors' => array(), 'runner' => false);
        if (self::$usenationalwebservices && ! self::$disablesyncwithnational) {
            $runnerstatus = get_config(self::BLOCK_NAME, self::BLOCK_RUNNER_KEY);
            $timecheck = time() - 60000; // 10 mins ago
            if (isset($runnerstatus) && ($runnerstatus > $timecheck)) {
                $results['runner'] = true;
                $results['errors'][] = 'sync is already running since '.date('Y-m-d h:i:s', $runnerstatus);
            } else {
                set_config(self::BLOCK_RUNNER_KEY, time(), self::BLOCK_NAME);
                try {
                     $localregsl = self::get_all_registrations(); // list
                    $nationalregsl = self::ws_get_students();
                    // put these into mapped lists so they can be more easily worked with and handled
                    $localregs = array();
                    if ( $localregsl) {
                        foreach ($localregsl as $reg) {
                            $id = self::make_reg_key($reg);
                            $localregs[$id] = $reg;
                        }
                    }
                    $nationalregs = array();
                    if ($nationalregsl) {
                        foreach ($nationalregsl as $reg) {
                            $id = self::make_reg_key($reg);
                            $nationalregs[$id] = $reg;
                        }
                    }

                    // create maps and sets of local and remote regs
                    // both contains the items that exist in both sets, only contains items from one set only
                    $nationalregsonly = array_diff_key($nationalregs, $localregs); // set
                    $localregsonly = array_diff_key($localregs, $nationalregs); // set
                    $nationalregsboth = array_diff_key($nationalregs, $localregsonly); // set
                    $localregsboth = array_diff_key($localregs, $nationalregsonly); // set

                    foreach ($localregsboth as $key => $localreg) {
                        // update if needed or just continue (push local or national or neither)
                        $nationalreg = array_key_exists($key, $nationalregsboth) ? $nationalregsboth[$key] : null;
                        if ($nationalreg != null) {
                            // compare these for diffs
                            /* * @noinspection PhpUndefinedFieldInspection */
                            if ($localreg->activated != $nationalreg->activated) {
                                try {
                                    /* * @noinspection PhpUndefinedFieldInspection */
                                    $nationalreg->activated = $localreg->activated;
                                    self::ws_save_clicker($nationalreg);
                                } catch (Exception $e) {
                                    // this is ok, we will continue anyway
                                    $msg = "Failed during national activate push all sync while syncing i>clicker registration ($nationalreg): $e";
                                    $results['errors'][] = $key;
                                    self::send_notifications($msg, $e);
                                    // log.warn(msg);
                                }
                            }
                        }
                    }

                    foreach ($nationalregsonly as $key => $nationalreg) {
                        try {
                            $reg = new stdClass();
                            $reg->clickerid = $nationalreg->clickerid;
                            $reg->ownerid = $nationalreg->ownerid;
                            $reg->activated = $nationalreg->activated;
                            $reg->fromnational = true;
                            self::save_registration($reg);
                        } catch (Exception $e) {
                            // this is ok, we will continue anyway
                            $msg = "Failed during local push all sync while syncing i>clicker registration ({$nationalreg->clickerid}): $e";
                            $results['errors'][] = $key;
                            self::send_notifications($msg, $e);
                            // log.warn(msg);
                        }
                    }

                    foreach ($localregsonly as $key => $localreg) {
                        try {
                            $reg = new stdClass();
                            $reg->clickerid = $localreg->clickerid;
                            $reg->ownerid = $localreg->ownerid;
                            $reg->activated = $localreg->activated;
                            $reg->fromnational = false;
                            self::ws_save_clicker($reg);
                        } catch (Exception $e) {
                            // this is ok, we will continue anyway
                            $msg = "Failed during national push all sync while syncing i>clicker registration ({$localreg->clickerid}): $e";
                            $results['errors'][] = $key;
                            self::send_notifications($msg, $e);
                            // log.warn(msg);
                        }
                    }
                    // log.info("Completed syncing "+total+" i>clicker registrations to ("+$localregsonly.size()+") and from ("+$nationalregsonly.size()+") national");
                    $results['local'] = count($localregsonly);
                    $results['national'] = count($nationalregsonly);
                    $results['total'] = count($localregsboth) + count($nationalregsonly);
                } catch (Exception $e) {
                    // reset the clicker runner
                    set_config(self::BLOCK_RUNNER_KEY, 0, self::BLOCK_NAME);
                    throw $e;
                }
            }
        } else {
            // disabled
            $results = false;
        }
        return $results;
    }

    private static function make_reg_key($reg) {
        return $reg->ownerid.':'.$reg->clickerid;
    }

    /**
     * Calls to the national webservices and gets all the student clicker registrations for the current domain
     * @return array of clicker registration objects
     */
    public static function ws_get_students() {
        $wsoperation = 'StudentsReport';
        $wsdomainurl = self::$domainurl; // 'http://epicurus.learningmate.com/';
        $arguments = array(
            'pVarUrl' => $wsdomainurl,
        );
        // $ws_soap_envelope = '<StudentsReport xmlns="http://www.iclicker.com/"> <pVarUrl>'.$wsdomainurl.'</pVarUrl> </StudentsReport>';
        $result = self::ws_soap_call($wsoperation, $arguments);
        $xml = $result['StudentsReportResult'];
        $regs = self::decode_ws_xml($xml);
        return $regs;
    }

    /**
     * Calls to the national webservices to get the clicker registrations for a specific student in this domain
     * @param string $username the username (not user id) for a student in this domain
     * @return array of clicker registration objects
     */
    public static function ws_get_student($username) {
        $wsoperation = 'SingleStudentReport';
        $wsdomainurl = self::$domainurl; // 'http://epicurus.learningmate.com/';
        $arguments = array(
            'pVarUrl' => $wsdomainurl,
            'pVarStudentId' => $username,
        );
        // $ws_soap_envelope = '<SingleStudentReport xmlns="http://www.iclicker.com/"> <pVarUrl>'.$wsdomainurl.'</pVarUrl> <pVarStudentId>';
        // $ws_soap_envelope .= $user_name.'</pVarStudentId> </SingleStudentReport>';
        $result = self::ws_soap_call($wsoperation, $arguments);
        $xml = $result['SingleStudentReportResult'];
        $regs = self::decode_ws_xml($xml);
        return $regs;
    }

    /**
     * Register a new clicker with the national webservices server for this domain,
     * will return the set of all registrations for the user who registered the clicker
     * @param object $clickerreg a clicker registration object, fields(ownerid, clickerid)
     * @return array of clicker registration objects
     */
    public static function ws_save_clicker($clickerreg) {
        $wsoperation = 'RegisterStudent';
        $regxml = self::encode_for_xml(self::encode_registration($clickerreg));
        $arguments = array(
            'pVarRegXml' => $regxml,
        );
        // $ws_soap_envelope = '<RegisterStudent xmlns="http://www.iclicker.com/"> <pVarRegXml>'.$regxml.'</pVarRegXml> </RegisterStudent>';
        $result = self::ws_soap_call($wsoperation, $arguments);
        $xml = $result['RegisterStudentResult'];
        $regs = self::decode_ws_xml($xml);
        return $regs;
    }

    /**
     * Handles the soap call to the national webservices server
     * @static
     * @param string $wsoperation the operation to perform (e.g. 'StudentsReport')
     * @param array $wsarguments the SOAP arguments to send
     * @return array the results of the SOAP call
     * @throws ClickerWebservicesException if the call fails
     */
    private static function ws_soap_call($wsoperation, $wsarguments) {
        $client = new SoapClient(self::$webservicesurl, array(
                        'login' => self::$webservicesusername,
                        'password' => self::$webservicespassword,
                        'encoding' => 'UTF-8',
                        'trace' => 1,
                        'exceptions' => 1,
                )
        );
        try {
            $result = $client->__soapCall($wsoperation, $wsarguments);
            if ($result instanceof SoapFault) {
                $msg = 'SoapFault occured: '.var_export($result, true);
                debugging($msg);
                throw new Exception($msg);
            }
        } catch (Exception $e) {
            $msg = 'Failure in the i>clicker webservices: '.var_export($client, true);
            debugging($msg);
            throw new ClickerWebservicesException($msg);
        }
        // convert result to an array
        $result = json_decode(json_encode($result), true);
        unset($client);
        return $result;

        /* won't work with .NET webservices
        $connection = soap_connect(self::NATIONAL_WS_URL);
        if (is_a($connection, 'SoapFault')) {
            throw new WebservicesException('Failure in SOAP connect: '.$connection);
        }
        $call = 'StudentsReport';
        $params = array();
        $result = soap_call($connection, $call, $params);
        */
        /* 1.9
        $soap_client = new soap_client(self::$webservicesurl, false);
        $err = $soap_client->getError();
        if ($err) {
            echo '<h2>SOAP constructor error:</h2><pre>' . $err . '</pre>';
            throw new WebservicesException('SOAP constructor error: '. $err);
        }
        $soap_client->setCredentials(self::$webservicesusername, self::$webservicespassword, 'basic');
        $soap_client->soap_defencoding = 'UTF-8';
        $soap_client->operation = $wsoperation;
        */
        /* won't work with .NET webservices
        $soap_client->setDefaultRpcParams(true);
        $header_action = new soapval('SOAPAction', 'string', 'http://www.iclicker.com/StudentsReport');
        $header_content_type = new soapval('Content-Type', 'string', 'text/xml; charset=utf-8');
        $soap_client->setHeaders(array($header_content_type, $header_action));
        $params = array('pVarUrl' => 'http://epicurus.learningmate.com/'); // $webservicesurl); // 'http://epicurus.learningmate.com/');
        $result = $soap_client->call('StudentsReport', $params, 'http://www.iclicker.com/', 'http://www.iclicker.com/StudentsReport');
        */
        /* * 1.9
        $soap_msg = $soap_client->serializeEnvelope($ws_soap_envelope);
        $result = $soap_client->send($soap_msg, 'http://www.iclicker.com/'.$wsoperation);
        if ($soap_client->fault) {
            // check for fault
            echo '<h2>SOAP Fault</h2><xmp>';
            var_export($result);
            echo '</xmp>';
            throw new WebservicesException('SOAP fault: '. $result);
        } else {
            // Check for errors
            $err = $soap_client->getError();
            if ($err) {
                // Display the error
                echo '<h2>SOAP Error</h2><pre>' . $err . '</pre>';
                throw new WebservicesException('SOAP error: '. $err);
            }
        }
         */
        /*
        echo "<xmp>";
        var_export($result);
        var_export($soap_client);
        echo "</xmp>";
        */
        // unset($soap_client);
        // return $result;
    }


    // XML support functions

    /**
     * encodes a string for inclusion in an xml document
     * @param string $value the value to encode
     * @return string the value with xml chars encoded and replaced
     */
    private static function encode_for_xml($value) {
        if ($value) {
            return htmlspecialchars($value, ENT_QUOTES, 'UTF-8');
        }
        return '';
    }

}

// load the config into the static vars from the global plugin config settings
$blockuniquename = iclicker_service::BLOCK_NAME;

// $block_iclicker_allow_sharing = get_config($blockuniquename, 'block_iclicker_allow_sharing');
$blockiclickerdisablealternateid = get_config($blockuniquename, 'block_iclicker_disable_alternateid');
$blockiclickerusenationalws = get_config($blockuniquename, 'block_iclicker_use_national_ws');
$blockiclickerdomainurl = get_config($blockuniquename, 'block_iclicker_domain_url');
$blockiclickerwebservicesurl = get_config($blockuniquename, 'block_iclicker_webservices_url');
$blockiclickerwebservicesusername = get_config($blockuniquename, 'block_iclicker_webservices_username');
$blockiclickerwebservicespassword = get_config($blockuniquename, 'block_iclicker_webservices_password');
$blockiclickerdisablesync = get_config($blockuniquename, 'block_iclicker_disable_sync');
$blockiclickermaxcoursestofetch = get_config($blockuniquename, 'block_iclicker_max_courses_fetched');
$blockiclickerssosharedkey = get_config($blockuniquename, 'block_iclicker_sso_shared_key');
$blockiclickerenableshortname = get_config($blockuniquename, 'block_iclicker_enable_shortname');

iclicker_service::$serverurl = $CFG->wwwroot;
if (!empty($blockiclickerdomainurl)) {
    iclicker_service::$domainurl = $blockiclickerdomainurl;
} else {
    iclicker_service::$domainurl = $CFG->wwwroot;
}
if (!empty($blockiclickerdisablealternateid)) {
    iclicker_service::$disablealternateid = true;
}
/* cannot change this for now
if (!empty($block_iclicker_allow_sharing)) {
    iclicker_service::$allowremotesharing = true;
}*/
if (!empty($blockiclickerusenationalws)) {
    iclicker_service::$usenationalwebservices = true;
}
if (!empty($blockiclickerwebservicesurl)) {
    iclicker_service::$webservicesurl = $blockiclickerwebservicesurl;
}
if (!empty($blockiclickerwebservicesusername)) {
    iclicker_service::$webservicesusername = $blockiclickerwebservicesusername;
}
if (!empty($blockiclickerwebservicespassword)) {
    iclicker_service::$webservicespassword = $blockiclickerwebservicespassword;
}
if (!empty($blockiclickerdisablesync)) {
    iclicker_service::$disablesyncwithnational = true;
}
if (!empty($blockiclickermaxcoursestofetch)
        && is_numeric($blockiclickermaxcoursestofetch)
        && $blockiclickermaxcoursestofetch > 0) {
    iclicker_service::$maxcoursestofetch = intval($blockiclickermaxcoursestofetch);
}
if (!empty($blockiclickerssosharedkey)) {
    iclicker_service::$blockiclickerssoenabled = true;
    iclicker_service::$blockiclickerssosharedkey = $blockiclickerssosharedkey;
}
if (!empty($blockiclickerenableshortname)) {
    iclicker_service::$enablecourseshortname = true;
}
